"""Test ZHA Gateway."""

import asyncio
from unittest.mock import AsyncMock, MagicMock, PropertyMock, call, patch

import pytest
from zigpy.application import ControllerApplication
from zigpy.profiles import zha
import zigpy.types
from zigpy.zcl.clusters import general, lighting
import zigpy.zdo.types
import zigpy.zdo.types as zdo_t
from zigpy.zdo.types import LogicalType, NodeDescriptor

from tests.common import (
    SIG_EP_INPUT,
    SIG_EP_OUTPUT,
    SIG_EP_PROFILE,
    SIG_EP_TYPE,
    create_mock_zigpy_device,
    get_entity,
    get_group_entity,
    join_zigpy_device,
)
from zha.application import Platform
from zha.application.const import (
    CONF_USE_THREAD,
    ZHA_GW_MSG,
    ZHA_GW_MSG_CONNECTION_LOST,
    RadioType,
)
from zha.application.gateway import (
    ConnectionLostEvent,
    DeviceJoinedDeviceInfo,
    DeviceJoinedEvent,
    DevicePairingStatus,
    Gateway,
    RawDeviceInitializedDeviceInfo,
    RawDeviceInitializedEvent,
)
from zha.application.helpers import ZHAData
from zha.application.platforms import GroupEntity
from zha.application.platforms.light.const import EFFECT_OFF, LightEntityFeature
from zha.zigbee.device import Device
from zha.zigbee.group import Group, GroupMemberReference

IEEE_GROUPABLE_DEVICE = "01:2d:6f:00:0a:90:69:e8"
IEEE_GROUPABLE_DEVICE2 = "02:2d:6f:00:0a:90:69:e8"


ZIGPY_DEVICE_BASIC = {
    1: {
        SIG_EP_INPUT: [general.Basic.cluster_id],
        SIG_EP_OUTPUT: [],
        SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
        SIG_EP_PROFILE: zha.PROFILE_ID,
    }
}


async def coordinator_mock(zha_gateway: Gateway) -> Device:
    """Test ZHA light platform."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.COLOR_DIMMABLE_LIGHT,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        ieee="00:15:8d:00:02:32:4f:32",
        nwk=0x0000,
        node_descriptor=zdo_t.NodeDescriptor(
            logical_type=zdo_t.LogicalType.Coordinator,
            complex_descriptor_available=0,
            user_descriptor_available=0,
            reserved=0,
            aps_flags=0,
            frequency_band=zdo_t.NodeDescriptor.FrequencyBand.Freq2400MHz,
            mac_capability_flags=(
                zdo_t.NodeDescriptor.MACCapabilityFlags.AlternatePanCoordinator
                | zdo_t.NodeDescriptor.MACCapabilityFlags.FullFunctionDevice
                | zdo_t.NodeDescriptor.MACCapabilityFlags.MainsPowered
                | zdo_t.NodeDescriptor.MACCapabilityFlags.RxOnWhenIdle
                | zdo_t.NodeDescriptor.MACCapabilityFlags.AllocateAddress
            ),
            manufacturer_code=43981,
            maximum_buffer_size=82,
            maximum_incoming_transfer_size=128,
            server_mask=11329,
            maximum_outgoing_transfer_size=128,
            descriptor_capability_field=zdo_t.NodeDescriptor.DescriptorCapability.NONE,
        ),
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    return zha_device


async def device_light_1_mock(zha_gateway: Gateway) -> Device:
    """Test ZHA light platform."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.OnOff.cluster_id,
                    general.LevelControl.cluster_id,
                    lighting.Color.cluster_id,
                    general.Groups.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.COLOR_DIMMABLE_LIGHT,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        ieee=IEEE_GROUPABLE_DEVICE,
        manufacturer="Philips",
        model="LWA004",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    return zha_device


async def device_light_2_mock(zha_gateway: Gateway) -> Device:
    """Test ZHA light platform."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.OnOff.cluster_id,
                    general.LevelControl.cluster_id,
                    lighting.Color.cluster_id,
                    general.Groups.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.COLOR_DIMMABLE_LIGHT,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        ieee=IEEE_GROUPABLE_DEVICE2,
        manufacturer="Sengled",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    return zha_device


async def test_device_left(zha_gateway: Gateway) -> None:
    """Device leaving the network should become unavailable."""
    zigpy_dev_basic = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE_BASIC)
    zha_dev_basic = await join_zigpy_device(zha_gateway, zigpy_dev_basic)
    assert zha_dev_basic.available is True

    zha_gateway.device_left(zigpy_dev_basic)
    await zha_gateway.async_block_till_done()
    assert zha_dev_basic.available is False
    assert zha_dev_basic.on_network is False


async def test_gateway_startup_failure(
    zha_data: ZHAData,
) -> None:
    """Test shutdown called when gateway init fails."""

    zha_gateway = await Gateway.async_from_config(zha_data)

    with (
        patch("zha.application.gateway.Gateway.load_devices", side_effect=Exception),
        pytest.raises(Exception),
    ):
        zha_gateway.shutdown = AsyncMock(wraps=zha_gateway.shutdown)
        await zha_gateway.async_initialize()
        await zha_gateway.async_block_till_done()

    assert zha_gateway.shutdown.await_count == 1


async def test_gateway_starts_entity_exception(
    zha_data: ZHAData,
    zigpy_app_controller: ControllerApplication,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test gateway starts when we fail to create an entity."""

    with (
        patch(
            "bellows.zigbee.application.ControllerApplication.new",
            return_value=zigpy_app_controller,
        ),
        patch(
            "bellows.zigbee.application.ControllerApplication",
            return_value=zigpy_app_controller,
        ),
        patch(
            "zha.application.platforms.sensor.DeviceCounterSensor.__init__",
            side_effect=Exception,
        ),
    ):
        zha_gateway = await Gateway.async_from_config(zha_data)
        await zha_gateway.async_initialize()
        await zha_gateway.async_block_till_done()
        await zha_gateway.async_initialize_devices_and_entities()

        assert "Failed to create entity" in caplog.text

        await zha_gateway.shutdown()


@pytest.mark.parametrize(("enabled", "await_count"), [(True, 1), (False, 0)])
async def test_mains_devices_startup_polling_config(
    zha_data: ZHAData,
    zigpy_app_controller: ControllerApplication,
    enabled: bool,
    await_count: int,
) -> None:
    """Test mains powered device startup polling config is respected."""

    with (
        patch(
            "bellows.zigbee.application.ControllerApplication.new",
            return_value=zigpy_app_controller,
        ),
        patch(
            "bellows.zigbee.application.ControllerApplication",
            return_value=zigpy_app_controller,
        ),
    ):
        zha_data.config.device_options.enable_mains_startup_polling = enabled
        zha_gateway = await Gateway.async_from_config(zha_data)
        zha_gateway.async_fetch_updated_state_mains = AsyncMock(
            wraps=zha_gateway.async_fetch_updated_state_mains
        )
        await zha_gateway.async_initialize()
        await zha_gateway.async_block_till_done()
        await zha_gateway.async_initialize_devices_and_entities()
        await zha_gateway.async_block_till_done()

        assert zha_gateway.async_fetch_updated_state_mains.await_count == await_count

        await zha_gateway.shutdown()
        await zha_gateway.async_block_till_done()


async def test_gateway_group_methods(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test creating a group with 2 members."""
    coordinator = await coordinator_mock(zha_gateway)
    zha_gateway.coordinator_zha_device = coordinator
    coordinator._zha_gateway = zha_gateway
    device_light_1 = await device_light_1_mock(zha_gateway)
    device_light_2 = await device_light_2_mock(zha_gateway)
    device_light_1._zha_gateway = zha_gateway
    device_light_2._zha_gateway = zha_gateway

    member_ieee_addresses = [device_light_1.ieee, device_light_2.ieee]
    members = [
        GroupMemberReference(ieee=device_light_1.ieee, endpoint_id=1),
        GroupMemberReference(ieee=device_light_2.ieee, endpoint_id=1),
    ]

    # test creating a group with 2 members
    zha_group: Group = await zha_gateway.async_create_zigpy_group("Test Group", members)
    await zha_gateway.async_block_till_done()

    assert zha_group is not None
    assert len(zha_group.members) == 2
    for member in zha_group.members:
        assert member.device.ieee in member_ieee_addresses
        assert member.group == zha_group
        assert member.endpoint is not None

    entity: GroupEntity | None = get_group_entity(zha_group, platform=Platform.LIGHT)
    assert entity is not None

    info = entity.info_object
    assert info.class_name == "LightGroup"
    assert info.platform == Platform.LIGHT
    assert info.unique_id == "light_zha_group_0x0002"
    assert info.fallback_name == "Test Group"
    assert info.group_id == zha_group.group_id
    assert info.supported_features == LightEntityFeature.TRANSITION
    assert info.min_mireds == 153
    assert info.max_mireds == 500
    assert info.effect_list == [EFFECT_OFF]

    device_1_light_entity = get_entity(device_light_1, platform=Platform.LIGHT)
    device_2_light_entity = get_entity(device_light_2, platform=Platform.LIGHT)
    assert device_1_light_entity != device_2_light_entity

    assert device_1_light_entity is not None
    assert device_2_light_entity is not None

    # test get group by name
    assert zha_group == zha_gateway.get_group(zha_group.name)

    # test removing a group
    await zha_gateway.async_remove_zigpy_group(zha_group.group_id)
    await zha_gateway.async_block_till_done()

    # we shouldn't have the group anymore
    assert zha_gateway.get_group(zha_group.name) is None

    # the group entity should be cleaned up
    with pytest.raises(KeyError):
        get_group_entity(zha_group, platform=Platform.LIGHT)

    # test creating a group with 1 member
    zha_group = await zha_gateway.async_create_zigpy_group(
        "Test Group", [GroupMemberReference(ieee=device_light_1.ieee, endpoint_id=1)]
    )
    await zha_gateway.async_block_till_done()

    assert zha_group is not None
    assert len(zha_group.members) == 1
    for member in zha_group.members:
        assert member.device.ieee in [device_light_1.ieee]

    # no entity should be created for a group with a single member
    with pytest.raises(KeyError):
        get_group_entity(zha_group, platform=Platform.LIGHT)

    with patch("zigpy.zcl.Cluster.request", side_effect=TimeoutError):
        await zha_group.members[0].async_remove_from_group()
        assert len(zha_group.members) == 1
        for member in zha_group.members:
            assert member.device.ieee in [device_light_1.ieee]

    await zha_gateway.async_remove_zigpy_group(23)
    await zha_gateway.async_block_till_done()
    assert "Group: 0x0017 could not be found" in caplog.text

    assert zha_gateway.get_group(zha_group.group_id) is not None
    assert zha_gateway.get_group(zha_group.group_id) == zha_group


async def test_gateway_create_group_with_id_without_id(zha_gateway: Gateway) -> None:
    """Test creating a group with a specific ID."""

    assert zha_gateway is not None
    coordinator = await coordinator_mock(zha_gateway)
    zha_gateway.coordinator_zha_device = coordinator
    device_light_1 = await device_light_1_mock(zha_gateway)
    coordinator._zha_gateway = zha_gateway
    device_light_1._zha_gateway = zha_gateway

    zha_group = await zha_gateway.async_create_zigpy_group(
        "Test Group",
        [GroupMemberReference(ieee=device_light_1.ieee, endpoint_id=1)],
        group_id=0x1234,
    )
    await zha_gateway.async_block_till_done()

    assert len(zha_group.members) == 1
    assert zha_group.members[0].device is device_light_1
    assert zha_group.group_id == 0x1234

    # create group with no group id passed in
    zha_group2 = await zha_gateway.async_create_zigpy_group(
        "Test Group2",
        [GroupMemberReference(ieee=device_light_1.ieee, endpoint_id=1)],
    )

    assert len(zha_group2.members) == 1
    assert zha_group2.members[0].device is device_light_1
    assert zha_group2.group_id == 0x0002

    # create group with no group id passed in
    zha_group3 = await zha_gateway.async_create_zigpy_group(
        "Test Group3",
        [GroupMemberReference(ieee=device_light_1.ieee, endpoint_id=1)],
    )

    assert len(zha_group3.members) == 1
    assert zha_group3.members[0].device is device_light_1
    assert zha_group3.group_id == 0x0003


async def test_remove_device_cleans_up_group_membership(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test group membership cleanup when removing a device in a group."""

    coordinator = await coordinator_mock(zha_gateway)
    zha_gateway.coordinator_zha_device = coordinator
    coordinator._zha_gateway = zha_gateway
    device_light_1 = await device_light_1_mock(zha_gateway)
    device_light_2 = await device_light_2_mock(zha_gateway)
    device_light_1._zha_gateway = zha_gateway
    device_light_2._zha_gateway = zha_gateway

    member_ieee_addresses = [device_light_1.ieee, device_light_2.ieee]
    members = [
        GroupMemberReference(ieee=device_light_1.ieee, endpoint_id=1),
        GroupMemberReference(ieee=device_light_2.ieee, endpoint_id=1),
    ]

    # test creating a group with 2 members
    zha_group: Group = await zha_gateway.async_create_zigpy_group("Test Group", members)
    await zha_gateway.async_block_till_done()

    assert zha_group is not None
    assert len(zha_group.members) == 2
    for member in zha_group.members:
        assert member.device.ieee in member_ieee_addresses
        assert member.group == zha_group
        assert member.endpoint is not None

    await zha_gateway.async_remove_device(coordinator.ieee)
    await zha_gateway.async_block_till_done()
    assert len(zha_group.members) == 2
    assert (
        f"Removing the active coordinator ({str(coordinator.ieee)}) is not allowed"
        in caplog.text
    )

    non_existent_ieee = zigpy.types.EUI64.convert("01:2d:6f:70:7a:40:79:e8")
    await zha_gateway.async_remove_device(non_existent_ieee)
    await zha_gateway.async_block_till_done()
    assert len(zha_group.members) == 2
    assert f"Device: {str(non_existent_ieee)} could not be found" in caplog.text

    await zha_gateway.async_remove_device(device_light_1.ieee)
    await zha_gateway.async_block_till_done()

    assert len(zha_group.members) == 1
    assert zha_group.members[0].device.ieee == device_light_2.ieee
    assert device_light_1.ieee not in zha_gateway.devices


@patch(
    "zha.application.gateway.Gateway.load_devices",
    AsyncMock(),
)
@patch(
    "zha.application.gateway.Gateway.load_groups",
    MagicMock(),
)
@pytest.mark.parametrize(
    ("device_path", "thread_state", "config_override"),
    [
        ("/dev/ttyUSB0", True, {}),
        ("socket://192.168.1.123:9999", False, {}),
        ("socket://192.168.1.123:9999", True, {"use_thread": True}),
    ],
)
async def test_gateway_initialize_bellows_thread(
    device_path: str,
    thread_state: bool,
    config_override: dict,
    zigpy_app_controller: ControllerApplication,
    zha_data: ZHAData,
) -> None:
    """Test ZHA disabling the UART thread when connecting to a TCP coordinator."""
    zha_data.config.coordinator_configuration.path = device_path
    zha_data.zigpy_config = config_override

    with patch(
        "bellows.zigbee.application.ControllerApplication.new",
        return_value=zigpy_app_controller,
    ) as mock_new:
        zha_gw = Gateway(zha_data)
        await zha_gw.async_initialize()
        assert (
            mock_new.mock_calls[-1].kwargs["config"].get(CONF_USE_THREAD, True)
            is thread_state
        )
        await zha_gw.shutdown()


@pytest.mark.parametrize("radio_concurrency", [1, 2, 8])
async def test_startup_concurrency_limit(
    radio_concurrency: int,
    zigpy_app_controller: ControllerApplication,
    zha_data: ZHAData,
):
    """Test ZHA gateway limits concurrency on startup."""
    zha_gw = Gateway(zha_data)

    with patch(
        "bellows.zigbee.application.ControllerApplication.new",
        return_value=zigpy_app_controller,
    ):
        await zha_gw.async_initialize()

    for i in range(50):
        zigpy_dev = create_mock_zigpy_device(
            zha_gw,
            {
                1: {
                    SIG_EP_INPUT: [
                        general.OnOff.cluster_id,
                        general.LevelControl.cluster_id,
                        lighting.Color.cluster_id,
                        general.Groups.cluster_id,
                    ],
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: zha.DeviceType.COLOR_DIMMABLE_LIGHT,
                    SIG_EP_PROFILE: zha.PROFILE_ID,
                }
            },
            ieee=f"11:22:33:44:{i:08x}",
            nwk=0x1234 + i,
        )
        zigpy_dev.node_desc.mac_capability_flags |= (
            zigpy.zdo.types.NodeDescriptor.MACCapabilityFlags.MainsPowered
        )

        zha_gw.get_or_create_device(zigpy_dev)

    # Keep track of request concurrency during initialization
    current_concurrency = 0
    concurrencies = []

    async def mock_send_packet(*args, **kwargs):  # pylint: disable=unused-argument
        """Mock send packet."""
        nonlocal current_concurrency

        current_concurrency += 1
        concurrencies.append(current_concurrency)

        await asyncio.sleep(0.001)

        current_concurrency -= 1
        concurrencies.append(current_concurrency)

    type(zha_gw).radio_concurrency = PropertyMock(return_value=radio_concurrency)
    assert zha_gw.radio_concurrency == radio_concurrency

    with patch(
        "zha.zigbee.device.Device.async_initialize",
        side_effect=mock_send_packet,
    ):
        await zha_gw.async_fetch_updated_state_mains()

    await zha_gw.shutdown()

    # Make sure concurrency was always limited
    assert current_concurrency == 0
    assert min(concurrencies) == 0

    if radio_concurrency > 1:
        assert 1 <= max(concurrencies) < zha_gw.radio_concurrency
    else:
        assert 1 == max(concurrencies) == zha_gw.radio_concurrency


async def test_gateway_device_removed(zha_gateway: Gateway) -> None:
    """Test ZHA device removal."""

    zigpy_dev_basic = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE_BASIC)
    zha_dev_basic = await join_zigpy_device(zha_gateway, zigpy_dev_basic)
    zha_gateway.device_removed(zigpy_dev_basic)
    await zha_gateway.async_block_till_done()
    assert zha_dev_basic.ieee not in zha_gateway.devices


async def test_gateway_device_initialized(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test ZHA device initialization."""

    zigpy_dev_basic = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE_BASIC)
    zha_gateway.async_device_initialized = AsyncMock(
        wraps=zha_gateway.async_device_initialized
    )
    zha_gateway.device_initialized(zigpy_dev_basic)
    await zha_gateway.async_block_till_done()

    assert (
        "Cancelling previous initialization task for device 00:0d:6f:00:0a:90:69:e7"
        not in caplog.text
    )

    assert zha_gateway.async_device_initialized.await_count == 1
    assert zha_gateway.async_device_initialized.await_args == call(zigpy_dev_basic)

    zha_gateway.async_device_initialized.reset_mock()

    # call 2x to make sure cancellation of the task happens
    zha_gateway.device_initialized(zigpy_dev_basic)
    assert (
        "Cancelling previous initialization task for device 00:0d:6f:00:0a:90:69:e7"
        not in caplog.text
    )
    zha_gateway.device_initialized(zigpy_dev_basic)
    await zha_gateway.async_block_till_done()

    assert (
        "Cancelling previous initialization task for device 00:0d:6f:00:0a:90:69:e7"
        in caplog.text
    )


def test_gateway_raw_device_initialized(
    zha_gateway: Gateway,
) -> None:
    """Test Zigpy raw device initialized."""

    zigpy_dev_basic = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE_BASIC)
    zha_gateway.emit = MagicMock(wraps=zha_gateway.emit)
    zha_gateway.raw_device_initialized(zigpy_dev_basic)

    assert zha_gateway.emit.call_count == 1
    assert zha_gateway.emit.call_args == call(
        "raw_device_initialized",
        RawDeviceInitializedEvent(
            device_info=RawDeviceInitializedDeviceInfo(
                ieee=zigpy.types.EUI64.convert("00:0d:6f:00:0a:90:69:e7"),
                nwk=0xB79C,
                pairing_status=DevicePairingStatus.INTERVIEW_COMPLETE,
                model="FakeModel",
                manufacturer="FakeManufacturer",
                signature={
                    "manufacturer": "FakeManufacturer",
                    "model": "FakeModel",
                    "node_desc": {
                        "logical_type": LogicalType.EndDevice,
                        "complex_descriptor_available": 0,
                        "user_descriptor_available": 0,
                        "reserved": 0,
                        "aps_flags": 0,
                        "frequency_band": NodeDescriptor.FrequencyBand.Freq2400MHz,
                        "mac_capability_flags": NodeDescriptor.MACCapabilityFlags.AllocateAddress,
                        "manufacturer_code": 4151,
                        "maximum_buffer_size": 127,
                        "maximum_incoming_transfer_size": 100,
                        "server_mask": 10752,
                        "maximum_outgoing_transfer_size": 100,
                        "descriptor_capability_field": NodeDescriptor.DescriptorCapability.NONE,
                    },
                    "endpoints": {
                        1: {
                            "profile_id": 260,
                            "device_type": zha.DeviceType.ON_OFF_SWITCH,
                            "input_clusters": [0],
                            "output_clusters": [],
                        }
                    },
                },
            ),
            event_type="zha_gateway_message",
            event="raw_device_initialized",
        ),
    )


def test_gateway_device_joined(
    zha_gateway: Gateway,
) -> None:
    """Test Zigpy raw device initialized."""

    zigpy_dev_basic = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE_BASIC)
    zha_gateway.emit = MagicMock(wraps=zha_gateway.emit)
    zha_gateway.device_joined(zigpy_dev_basic)

    assert zha_gateway.emit.call_count == 1
    assert zha_gateway.emit.call_args == call(
        "device_joined",
        DeviceJoinedEvent(
            device_info=DeviceJoinedDeviceInfo(
                ieee=zigpy.types.EUI64.convert("00:0d:6f:00:0a:90:69:e7"),
                nwk=0xB79C,
                pairing_status=DevicePairingStatus.PAIRED,
            )
        ),
    )


def test_gateway_connection_lost(zha_gateway: Gateway) -> None:
    """Test Zigpy raw device initialized."""

    exception = Exception("Test exception")
    zha_gateway.emit = MagicMock(wraps=zha_gateway.emit)
    zha_gateway.connection_lost(exception)

    assert zha_gateway.emit.call_count == 1
    assert zha_gateway.emit.call_args == call(
        ZHA_GW_MSG_CONNECTION_LOST,
        ConnectionLostEvent(
            exception=exception,
            event=ZHA_GW_MSG_CONNECTION_LOST,
            event_type=ZHA_GW_MSG,
        ),
    )


async def test_pollers_skip(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test pollers skip when they should."""

    assert "Global updater interval skipped" not in caplog.text
    assert "Device availability checker interval skipped" not in caplog.text

    assert zha_gateway.config.allow_polling is True
    zha_gateway.config.allow_polling = False
    assert zha_gateway.config.allow_polling is False

    sleep_time = max(
        zha_gateway.global_updater.__polling_interval,
        zha_gateway._device_availability_checker.__polling_interval,
    )
    sleep_time += 2

    await asyncio.sleep(sleep_time)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert "Global updater interval skipped" in caplog.text
    assert "Device availability checker interval skipped" in caplog.text


async def test_global_updater_guards(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test global updater guards."""

    already_registered = (
        "listener already registered with global updater - nothing to register"
    )
    not_registered = "listener not registered with global updater - nothing to remove"

    assert already_registered not in caplog.text
    assert not_registered not in caplog.text

    def listener():
        pass

    zha_gateway.global_updater.register_update_listener(listener)

    assert already_registered not in caplog.text

    zha_gateway.global_updater.register_update_listener(listener)

    assert already_registered in caplog.text

    zha_gateway.global_updater.remove_update_listener(listener)

    assert not_registered not in caplog.text

    zha_gateway.global_updater.remove_update_listener(listener)

    assert not_registered in caplog.text


async def test_gateway_handle_message(
    zha_gateway: Gateway,
) -> None:
    """Test handle message."""

    zigpy_dev_basic = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE_BASIC)
    zha_dev_basic = await join_zigpy_device(zha_gateway, zigpy_dev_basic)
    assert zha_dev_basic.available is True
    assert zha_dev_basic.on_network is True

    zha_dev_basic.on_network = False

    assert zha_dev_basic.available is False
    assert zha_dev_basic.on_network is False

    zha_gateway.handle_message(
        zha_dev_basic.device,
        zha.PROFILE_ID,
        general.Basic.cluster_id,
        1,
        1,
        b"",
    )

    assert zha_dev_basic.available is True
    assert zha_dev_basic.on_network is True


def test_radio_type():
    """Test radio type."""

    assert RadioType.list() == [
        "EZSP = Silicon Labs EmberZNet protocol: Elelabs, HUSBZB-1, Telegesis",
        "ZNP = Texas Instruments Z-Stack ZNP protocol: CC253x, CC26x2, CC13x2",
        "deCONZ = dresden elektronik deCONZ protocol: ConBee I/II, RaspBee I/II",
        "ZiGate = ZiGate Zigbee radios: PiZiGate, ZiGate USB-TTL, ZiGate WiFi",
        "XBee = Digi XBee Zigbee radios: Digi XBee Series 2, 2C, 3",
    ]

    assert (
        RadioType.get_by_description(
            "EZSP = Silicon Labs EmberZNet protocol: Elelabs, HUSBZB-1, Telegesis"
        )
        == RadioType.ezsp
    )

    assert RadioType.ezsp.description == (
        "EZSP = Silicon Labs EmberZNet protocol: Elelabs, HUSBZB-1, Telegesis"
    )

    with pytest.raises(ValueError):
        RadioType.get_by_description("Invalid description")


async def test_gateway_network_scan(zha_gateway: Gateway) -> None:
    """Test gateway network_scan method."""
    channels = zigpy.types.Channels.from_channel_list([11, 15, 20, 25])
    duration_exp = 3

    # Create mock network beacons
    mock_beacon1 = zigpy.types.NetworkBeacon(
        src=0x1234,
        channel=11,
        pan_id=0xABCD,
        extended_pan_id=zigpy.types.ExtendedPanId.convert("00:11:22:33:44:55:66:77"),
        permit_joining=True,
        router_capacity=True,
        device_capacity=False,
        lqi=150,
        rssi=-45,
        nwk_update_id=0,
        stack_profile=2,
        protocol_version=2,
    )

    mock_beacon2 = zigpy.types.NetworkBeacon(
        src=None,
        channel=15,
        pan_id=0xEF01,
        extended_pan_id=zigpy.types.ExtendedPanId.convert("aa:bb:cc:dd:ee:ff:00:11"),
        permit_joining=False,
        router_capacity=False,
        device_capacity=True,
        lqi=100,
        rssi=-60,
        nwk_update_id=1,
        stack_profile=2,
        protocol_version=2,
    )

    # Mock the application_controller's network_scan method
    async def mock_network_scan(channels, duration_exp):
        yield mock_beacon1
        yield mock_beacon2

    with patch.object(
        zha_gateway.application_controller,
        "network_scan",
        side_effect=mock_network_scan,
    ) as mock_scan:
        beacons = []
        async for beacon in zha_gateway.network_scan(channels, duration_exp):
            beacons.append(beacon)

        assert beacons == [mock_beacon1, mock_beacon2]
        assert mock_scan.mock_calls == [
            call(channels=channels, duration_exp=duration_exp)
        ]


async def test_gateway_energy_scan(zha_gateway: Gateway) -> None:
    """Test gateway energy_scan method."""
    channels = zigpy.types.Channels.from_channel_list([11, 15, 20, 25])
    expected_results = {
        11: 45.5,
        15: 78.2,
        20: 32.1,
        25: 91.7,
    }

    with patch.object(
        zha_gateway.application_controller, "energy_scan", return_value=expected_results
    ) as mock_scan:
        results = await zha_gateway.energy_scan(channels, 4, 1)

        assert results == expected_results
        assert mock_scan.mock_calls == [
            call(channels=channels, duration_exp=4, count=1)
        ]



"""Test ZHA Core cluster_handlers."""

# pylint:disable=redefined-outer-name,too-many-lines

from __future__ import annotations

import logging
import math
from typing import TYPE_CHECKING, Any
from unittest import mock
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from zhaquirks.centralite.cl_3130 import CentraLite3130
from zhaquirks.xiaomi.aqara.sensor_switch_aq3 import BUTTON_DEVICE_TYPE, SwitchAQ3
from zigpy.device import Device as ZigpyDevice
from zigpy.endpoint import Endpoint as ZigpyEndpoint
import zigpy.profiles.zha
from zigpy.quirks import _DEVICE_REGISTRY
import zigpy.types as t
from zigpy.zcl import foundation
import zigpy.zcl.clusters
from zigpy.zcl.clusters import CLUSTERS_BY_ID
from zigpy.zcl.clusters.general import (
    Basic,
    Identify,
    LevelControl,
    MultistateInput,
    OnOff,
    Ota,
    PollControl,
    PowerConfiguration,
)
from zigpy.zcl.clusters.homeautomation import Diagnostic
from zigpy.zcl.clusters.measurement import TemperatureMeasurement
import zigpy.zdo.types as zdo_t

from tests.common import (
    SIG_EP_INPUT,
    SIG_EP_OUTPUT,
    SIG_EP_PROFILE,
    SIG_EP_TYPE,
    create_mock_zigpy_device,
    get_entity,
    join_zigpy_device,
    send_attributes_report,
    zigpy_device_from_json,
)
from zha.application import Platform
from zha.application.const import ATTR_QUIRK_ID
from zha.application.gateway import Gateway
from zha.application.platforms.button import IdentifyButton
from zha.exceptions import ZHAException
from zha.zigbee.cluster_handlers import (
    AttrReportConfig,
    ClientClusterHandler,
    ClusterHandler,
    ClusterHandlerStatus,
    parse_and_log_command,
    retry_request,
)
from zha.zigbee.cluster_handlers.const import (
    CLUSTER_HANDLER_COLOR,
    CLUSTER_HANDLER_LEVEL,
    CLUSTER_HANDLER_ON_OFF,
)
from zha.zigbee.cluster_handlers.lighting import ColorClusterHandler
from zha.zigbee.cluster_handlers.lightlink import LightLinkClusterHandler
from zha.zigbee.cluster_handlers.registries import (
    CLIENT_CLUSTER_HANDLER_REGISTRY,
    CLUSTER_HANDLER_REGISTRY,
)
from zha.zigbee.device import Device
from zha.zigbee.endpoint import Endpoint


@pytest.fixture
def ieee():
    """IEEE fixture."""
    return t.EUI64.deserialize(b"ieeeaddr")[0]


@pytest.fixture
def nwk():
    """NWK fixture."""
    return t.NWK(0xBEEF)


def zigpy_coordinator_device_mock(zha_gateway: Gateway) -> ZigpyDevice:
    """Coordinator device fixture."""

    coordinator = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [0x1000],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: 0x1234,
                SIG_EP_PROFILE: 0x0104,
            }
        },
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
        nwk=0x0000,
    )
    coordinator.add_to_group = AsyncMock(return_value=[0])
    return coordinator


def endpoint_mock(zigpy_coordinator_device: ZigpyDevice) -> Endpoint:
    """Endpoint cluster_handlers fixture."""
    endpoint_mock = mock.MagicMock(spec_set=Endpoint)
    endpoint_mock.zigpy_endpoint.device.application.get_device.return_value = (
        zigpy_coordinator_device
    )
    endpoint_mock.device.skip_configuration = False
    endpoint_mock.id = 1
    return endpoint_mock


@pytest.mark.parametrize(
    ("cluster_id", "bind_count", "attrs"),
    [
        (zigpy.zcl.clusters.general.Basic.cluster_id, 0, set()),
        (
            zigpy.zcl.clusters.general.PowerConfiguration.cluster_id,
            1,
            {"battery_voltage", "battery_percentage_remaining"},
        ),
        (
            zigpy.zcl.clusters.general.DeviceTemperature.cluster_id,
            1,
            {"current_temperature"},
        ),
        (zigpy.zcl.clusters.general.Identify.cluster_id, 0, set()),
        (zigpy.zcl.clusters.general.Groups.cluster_id, 0, set()),
        (zigpy.zcl.clusters.general.Scenes.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.OnOff.cluster_id, 1, {"on_off"}),
        (zigpy.zcl.clusters.general.OnOffConfiguration.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.LevelControl.cluster_id, 1, {"current_level"}),
        (zigpy.zcl.clusters.general.Alarms.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.AnalogInput.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.AnalogOutput.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.AnalogValue.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.AnalogOutput.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.BinaryOutput.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.BinaryValue.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.MultistateInput.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.MultistateOutput.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.MultistateValue.cluster_id, 1, {"present_value"}),
        (zigpy.zcl.clusters.general.Commissioning.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.Partition.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.Ota.cluster_id, 0, set()),
        (zigpy.zcl.clusters.general.PowerProfile.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.ApplianceControl.cluster_id, 1, set()),
        (zigpy.zcl.clusters.general.GreenPowerProxy.cluster_id, 0, set()),
        (zigpy.zcl.clusters.closures.DoorLock.cluster_id, 1, {"lock_state"}),
        (
            zigpy.zcl.clusters.hvac.Thermostat.cluster_id,
            1,
            {
                "local_temperature",
                "occupied_cooling_setpoint",
                "occupied_heating_setpoint",
                "unoccupied_cooling_setpoint",
                "unoccupied_heating_setpoint",
                "running_mode",
                "running_state",
                "system_mode",
                "occupancy",
                "pi_cooling_demand",
                "pi_heating_demand",
            },
        ),
        (zigpy.zcl.clusters.hvac.Fan.cluster_id, 1, {"fan_mode"}),
        (
            zigpy.zcl.clusters.lighting.Color.cluster_id,
            1,
            {
                "current_x",
                "current_y",
                "color_temperature",
            },
        ),
        (
            zigpy.zcl.clusters.measurement.IlluminanceMeasurement.cluster_id,
            1,
            {"measured_value"},
        ),
        (
            zigpy.zcl.clusters.measurement.IlluminanceLevelSensing.cluster_id,
            1,
            {"level_status"},
        ),
        (
            zigpy.zcl.clusters.measurement.TemperatureMeasurement.cluster_id,
            1,
            {"measured_value"},
        ),
        (
            zigpy.zcl.clusters.measurement.PressureMeasurement.cluster_id,
            1,
            {"measured_value"},
        ),
        (
            zigpy.zcl.clusters.measurement.FlowMeasurement.cluster_id,
            1,
            {"measured_value"},
        ),
        (
            zigpy.zcl.clusters.measurement.RelativeHumidity.cluster_id,
            1,
            {"measured_value"},
        ),
        (zigpy.zcl.clusters.measurement.OccupancySensing.cluster_id, 1, {"occupancy"}),
        (
            zigpy.zcl.clusters.smartenergy.Metering.cluster_id,
            1,
            {
                "instantaneous_demand",
                "current_summ_delivered",
                "current_tier1_summ_delivered",
                "current_tier2_summ_delivered",
                "current_tier3_summ_delivered",
                "current_tier4_summ_delivered",
                "current_tier5_summ_delivered",
                "current_tier6_summ_delivered",
                "current_summ_received",
                "status",
            },
        ),
        (
            zigpy.zcl.clusters.homeautomation.ElectricalMeasurement.cluster_id,
            1,
            {
                "ac_frequency",
                "ac_voltage_divisor",
                "ac_current_divisor",
                "ac_power_divisor",
                "ac_voltage_multiplier",
                "ac_power_multiplier",
                "power_divisor",
                "power_multiplier",
                "ac_current_multiplier",
                "ac_frequency",
                "active_power",
                "active_power_ph_b",
                "active_power_ph_c",
                "total_active_power",
                "apparent_power",
                "rms_current",
                "rms_current_ph_b",
                "rms_current_ph_c",
                "rms_voltage",
                "rms_voltage_ph_b",
                "rms_voltage_ph_c",
            },
        ),
    ],
)
async def test_in_cluster_handler_config(
    cluster_id, bind_count, attrs, zha_gateway: Gateway
) -> None:
    """Test ZHA core cluster handler configuration for input clusters."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_INPUT: [cluster_id], SIG_EP_OUTPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster = zigpy_dev.endpoints[1].in_clusters[cluster_id]
    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        cluster_id, {None, ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    assert cluster_handler.status == ClusterHandlerStatus.CREATED

    if TYPE_CHECKING:
        # Workaround for https://github.com/python/mypy/issues/9005
        assert cluster_handler.status != ClusterHandlerStatus.CREATED

    await cluster_handler.async_configure()

    assert cluster_handler.status == ClusterHandlerStatus.CONFIGURED

    assert cluster.bind.call_count == bind_count

    reported_attrs = set()

    for mock_call in cluster.configure_reporting_multiple.mock_calls:
        reported_attrs.update(mock_call.args[0].keys())

    assert attrs == reported_attrs
    assert cluster.configure_reporting.call_count == 0
    assert cluster.configure_reporting_multiple.call_count == math.ceil(len(attrs) / 3)


async def test_cluster_handler_bind_error(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test ZHA core cluster handler bind error."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_INPUT: [OnOff.cluster_id], SIG_EP_OUTPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster: zigpy.zcl.Cluster = zigpy_dev.endpoints[1].in_clusters[OnOff.cluster_id]
    cluster.bind.side_effect = zigpy.exceptions.ZigbeeException

    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        OnOff.cluster_id, {None, ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    await cluster_handler.async_configure()

    assert cluster.bind.await_count == 3
    assert cluster.configure_reporting.await_count == 0
    assert f"Failed to bind '{cluster.ep_attribute}' cluster:" in caplog.text


async def test_cluster_handler_configure_reporting_error(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test ZHA core cluster handler configure reporting error."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_INPUT: [OnOff.cluster_id], SIG_EP_OUTPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster: zigpy.zcl.Cluster = zigpy_dev.endpoints[1].in_clusters[OnOff.cluster_id]
    cluster.configure_reporting_multiple.side_effect = zigpy.exceptions.ZigbeeException

    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        OnOff.cluster_id, {None, ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    await cluster_handler.async_configure()

    assert cluster.bind.await_count == 1
    assert cluster.configure_reporting_multiple.await_count == 3
    assert f"failed to set reporting on '{cluster.ep_attribute}' cluster" in caplog.text


async def test_write_attributes_safe_key_error(zha_gateway: Gateway) -> None:
    """Test ZHA core cluster handler write attributes safe key error."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_INPUT: [OnOff.cluster_id], SIG_EP_OUTPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster: zigpy.zcl.Cluster = zigpy_dev.endpoints[1].in_clusters[OnOff.cluster_id]
    cluster.write_attributes = AsyncMock(
        return_value=[
            foundation.WriteAttributesResponse.deserialize(b"\x01\x00\x00")[0]
        ]
    )

    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        OnOff.cluster_id, {None, ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    with pytest.raises(ZHAException, match="Failed to write attribute on_off=bar"):
        await cluster_handler.write_attributes_safe(
            {OnOff.AttributeDefs.on_off.name: "bar"}
        )


async def test_get_attributes_error(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test ZHA core cluster handler get attributes timeout error."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_INPUT: [OnOff.cluster_id], SIG_EP_OUTPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster: zigpy.zcl.Cluster = zigpy_dev.endpoints[1].in_clusters[OnOff.cluster_id]
    cluster.read_attributes.side_effect = zigpy.exceptions.ZigbeeException

    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        OnOff.cluster_id, {None, ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    await cluster_handler.get_attributes(["foo"])

    assert (
        f"failed to get attributes '['foo']' on '{OnOff.ep_attribute}' cluster"
        in caplog.text
    )


async def test_get_attributes_error_raises(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test ZHA core cluster handler get attributes timeout error."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_INPUT: [OnOff.cluster_id], SIG_EP_OUTPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster: zigpy.zcl.Cluster = zigpy_dev.endpoints[1].in_clusters[OnOff.cluster_id]
    cluster.read_attributes.side_effect = zigpy.exceptions.ZigbeeException

    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        OnOff.cluster_id, {None, ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    with pytest.raises(zigpy.exceptions.ZigbeeException):
        await cluster_handler._get_attributes(True, ["foo"])

    assert (
        f"failed to get attributes '['foo']' on '{OnOff.ep_attribute}' cluster"
        in caplog.text
    )


@pytest.mark.parametrize(
    ("cluster_id", "bind_count"),
    [
        (0x0000, 0),
        (0x0001, 1),
        (0x0002, 1),
        (0x0003, 0),
        (0x0004, 0),
        (0x0005, 1),
        (0x0006, 1),
        (0x0007, 1),
        (0x0008, 1),
        (0x0009, 1),
        (0x0015, 1),
        (0x0016, 1),
        (0x0019, 0),
        (0x001A, 1),
        (0x001B, 1),
        (0x0020, 1),
        (0x0021, 0),
        (0x0101, 1),
        (0x0202, 1),
        (0x0300, 1),
        (0x0400, 1),
        (0x0402, 1),
        (0x0403, 1),
        (0x0405, 1),
        (0x0406, 1),
        (0x0702, 1),
        (0x0B04, 1),
    ],
)
async def test_out_cluster_handler_config(
    cluster_id: int,
    bind_count: int,
    zha_gateway: Gateway,
) -> None:
    """Test ZHA core cluster handler configuration for output clusters."""
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {1: {SIG_EP_OUTPUT: [cluster_id], SIG_EP_INPUT: [], SIG_EP_TYPE: 0x1234}},
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )

    cluster = zigpy_dev.endpoints[1].out_clusters[cluster_id]
    cluster.bind_only = True
    cluster_handler_class: type[ClusterHandler] | None = CLUSTER_HANDLER_REGISTRY.get(
        cluster_id, {None: ClusterHandler}
    ).get(None)
    assert cluster_handler_class is not None
    cluster_handler = cluster_handler_class(cluster, endpoint)

    await cluster_handler.async_configure()

    assert cluster.bind.call_count == bind_count
    assert cluster.configure_reporting.call_count == 0


def test_cluster_handler_registry() -> None:
    """Test ZIGBEE cluster handler Registry."""

    # get all quirk ID from zigpy quirks registry
    cluster_quirk_id_map: dict[int, set[str | None]] = {}
    for cluster_id in CLUSTERS_BY_ID:
        cluster_quirk_id_map[cluster_id] = {None}

    # loop over custom clusters in v2 quirks registry
    for quirks in _DEVICE_REGISTRY.registry_v2.values():
        for quirk_reg_entry in quirks:
            # get standalone adds_metadata and adds_metadata from replaces_metadata
            all_metadata = set(quirk_reg_entry.adds_metadata) | {
                rm.add for rm in quirk_reg_entry.replaces_metadata
            }
            for metadata in all_metadata:
                cluster_quirk_id_map[metadata.cluster.cluster_id] = {None}

    # loop over custom clusters in v1 quirks registry
    for manufacturer in _DEVICE_REGISTRY.registry_v1.values():
        for model_quirk_list in manufacturer.values():
            for quirk in model_quirk_list:
                quirk_id = getattr(quirk, ATTR_QUIRK_ID, None)
                device_description: dict[str, dict[str, Any]] = getattr(
                    quirk, "replacement", None
                ) or getattr(quirk, "signature", None)

                for endpoint in device_description["endpoints"].values():
                    cluster_ids = set()
                    if "input_clusters" in endpoint:
                        cluster_ids.update(endpoint["input_clusters"])
                    if "output_clusters" in endpoint:
                        cluster_ids.update(endpoint["output_clusters"])
                    for cluster_id in cluster_ids:
                        if not isinstance(cluster_id, int):
                            cluster_id = cluster_id.cluster_id
                        if cluster_id not in cluster_quirk_id_map:
                            cluster_quirk_id_map[cluster_id] = {None}
                        cluster_quirk_id_map[cluster_id].add(quirk_id)

    for (
        cluster_id,
        cluster_handler_classes,
    ) in CLUSTER_HANDLER_REGISTRY.items():
        assert isinstance(cluster_id, int)
        assert 0 <= cluster_id <= 0xFFFF
        assert cluster_id in cluster_quirk_id_map
        assert isinstance(cluster_handler_classes, dict)
        for quirk_id, cluster_handler in cluster_handler_classes.items():
            assert quirk_id is None or isinstance(quirk_id, str)
            assert issubclass(cluster_handler, ClusterHandler)
            assert quirk_id in cluster_quirk_id_map[cluster_id]


def test_epch_unclaimed_cluster_handlers(cluster_handler) -> None:
    """Test unclaimed cluster handlers."""

    ch_1 = cluster_handler(CLUSTER_HANDLER_ON_OFF, 6)
    ch_2 = cluster_handler(CLUSTER_HANDLER_LEVEL, 8)
    ch_3 = cluster_handler(CLUSTER_HANDLER_COLOR, 768)

    mock_dev = mock.MagicMock(spec=Device)
    mock_dev.unique_id = "00:11:22:33:44:55:66:77"

    ep_cluster_handlers = Endpoint(mock.MagicMock(spec_set=ZigpyEndpoint), mock_dev)
    all_cluster_handlers = {ch_1.id: ch_1, ch_2.id: ch_2, ch_3.id: ch_3}
    with mock.patch.dict(
        ep_cluster_handlers.all_cluster_handlers, all_cluster_handlers, clear=True
    ):
        available = ep_cluster_handlers.unclaimed_cluster_handlers()
        assert ch_1 in available
        assert ch_2 in available
        assert ch_3 in available

        ep_cluster_handlers.claimed_cluster_handlers[ch_2.id] = ch_2
        available = ep_cluster_handlers.unclaimed_cluster_handlers()
        assert ch_1 in available
        assert ch_2 not in available
        assert ch_3 in available

        ep_cluster_handlers.claimed_cluster_handlers[ch_1.id] = ch_1
        available = ep_cluster_handlers.unclaimed_cluster_handlers()
        assert ch_1 not in available
        assert ch_2 not in available
        assert ch_3 in available

        ep_cluster_handlers.claimed_cluster_handlers[ch_3.id] = ch_3
        available = ep_cluster_handlers.unclaimed_cluster_handlers()
        assert ch_1 not in available
        assert ch_2 not in available
        assert ch_3 not in available


def test_epch_claim_cluster_handlers(cluster_handler) -> None:
    """Test cluster handler claiming."""

    ch_1 = cluster_handler(CLUSTER_HANDLER_ON_OFF, 6)
    ch_2 = cluster_handler(CLUSTER_HANDLER_LEVEL, 8)
    ch_3 = cluster_handler(CLUSTER_HANDLER_COLOR, 768)

    mock_dev = mock.MagicMock(spec=Device)
    mock_dev.unique_id = "00:11:22:33:44:55:66:77"

    ep_cluster_handlers = Endpoint(mock.MagicMock(spec_set=ZigpyEndpoint), mock_dev)
    all_cluster_handlers = {ch_1.id: ch_1, ch_2.id: ch_2, ch_3.id: ch_3}
    with mock.patch.dict(
        ep_cluster_handlers.all_cluster_handlers, all_cluster_handlers, clear=True
    ):
        assert ch_1.id not in ep_cluster_handlers.claimed_cluster_handlers
        assert ch_2.id not in ep_cluster_handlers.claimed_cluster_handlers
        assert ch_3.id not in ep_cluster_handlers.claimed_cluster_handlers

        ep_cluster_handlers.claim_cluster_handlers([ch_2])
        assert ch_1.id not in ep_cluster_handlers.claimed_cluster_handlers
        assert ch_2.id in ep_cluster_handlers.claimed_cluster_handlers
        assert ep_cluster_handlers.claimed_cluster_handlers[ch_2.id] is ch_2
        assert ch_3.id not in ep_cluster_handlers.claimed_cluster_handlers

        ep_cluster_handlers.claim_cluster_handlers([ch_3, ch_1])
        assert ch_1.id in ep_cluster_handlers.claimed_cluster_handlers
        assert ep_cluster_handlers.claimed_cluster_handlers[ch_1.id] is ch_1
        assert ch_2.id in ep_cluster_handlers.claimed_cluster_handlers
        assert ep_cluster_handlers.claimed_cluster_handlers[ch_2.id] is ch_2
        assert ch_3.id in ep_cluster_handlers.claimed_cluster_handlers
        assert ep_cluster_handlers.claimed_cluster_handlers[ch_3.id] is ch_3
        assert "1:0x0300" in ep_cluster_handlers.claimed_cluster_handlers


@mock.patch("zha.zigbee.endpoint.Endpoint.add_client_cluster_handlers")
@mock.patch(
    "zha.application.discovery.ENDPOINT_PROBE.discover_entities",
    mock.MagicMock(),
)
async def test_ep_cluster_handlers_all_cluster_handlers(
    m1,  # pylint: disable=unused-argument
    zha_gateway: Gateway,
) -> None:
    """Test Endpointcluster_handlers adding all cluster_handlers."""
    zha_device = await join_zigpy_device(
        zha_gateway,
        create_mock_zigpy_device(
            zha_gateway,
            {
                1: {
                    SIG_EP_INPUT: [0, 1, 6, 8],
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
                    SIG_EP_PROFILE: 0x0104,
                },
                2: {
                    SIG_EP_INPUT: [0, 1, 6, 8, 768],
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x0000,
                    SIG_EP_PROFILE: 0x0104,
                },
            },
        ),
    )
    assert "1:0x0000" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0001" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0006" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0008" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0300" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0000" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0001" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0006" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0008" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0300" not in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0000" not in zha_device._endpoints[2].all_cluster_handlers
    assert "1:0x0001" not in zha_device._endpoints[2].all_cluster_handlers
    assert "1:0x0006" not in zha_device._endpoints[2].all_cluster_handlers
    assert "1:0x0008" not in zha_device._endpoints[2].all_cluster_handlers
    assert "1:0x0300" not in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0000" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0001" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0006" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0008" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0300" in zha_device._endpoints[2].all_cluster_handlers


@mock.patch("zha.zigbee.endpoint.Endpoint.add_client_cluster_handlers")
@mock.patch(
    "zha.application.discovery.ENDPOINT_PROBE.discover_entities",
    mock.MagicMock(),
)
async def test_cluster_handler_power_config(
    m1,  # pylint: disable=unused-argument
    zha_gateway: Gateway,
) -> None:
    """Test that cluster_handlers only get a single power cluster_handler."""
    in_clusters = [0, 1, 6, 8]
    zha_device: Device = await join_zigpy_device(
        zha_gateway,
        create_mock_zigpy_device(
            zha_gateway,
            endpoints={
                1: {
                    SIG_EP_INPUT: in_clusters,
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x0000,
                    SIG_EP_PROFILE: 0x0104,
                },
                2: {
                    SIG_EP_INPUT: [*in_clusters, 768],
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x0000,
                    SIG_EP_PROFILE: 0x0104,
                },
            },
            ieee="01:2d:6f:00:0a:90:69:e8",
        ),
    )
    assert "1:0x0000" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0001" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0006" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0008" in zha_device._endpoints[1].all_cluster_handlers
    assert "1:0x0300" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0000" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0001" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0006" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0008" in zha_device._endpoints[2].all_cluster_handlers
    assert "2:0x0300" in zha_device._endpoints[2].all_cluster_handlers

    zha_device = await join_zigpy_device(
        zha_gateway,
        create_mock_zigpy_device(
            zha_gateway,
            endpoints={
                1: {
                    SIG_EP_INPUT: [],
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x0000,
                    SIG_EP_PROFILE: 0x0104,
                },
                2: {
                    SIG_EP_INPUT: in_clusters,
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x0000,
                    SIG_EP_PROFILE: 0x0104,
                },
            },
            ieee="02:2d:6f:00:0a:90:69:e8",
        ),
    )
    assert "1:0x0001" not in zha_device._endpoints[1].all_cluster_handlers
    assert "2:0x0001" in zha_device._endpoints[2].all_cluster_handlers

    zha_device = await join_zigpy_device(
        zha_gateway,
        create_mock_zigpy_device(
            zha_gateway,
            endpoints={
                2: {
                    SIG_EP_INPUT: in_clusters,
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x0000,
                    SIG_EP_PROFILE: 0x0104,
                }
            },
            ieee="03:2d:6f:00:0a:90:69:e8",
        ),
    )
    assert "2:0x0001" in zha_device._endpoints[2].all_cluster_handlers


async def test_ep_cluster_handlers_configure(cluster_handler) -> None:
    """Test unclaimed cluster handlers."""

    ch_1 = cluster_handler(CLUSTER_HANDLER_ON_OFF, 6)
    ch_2 = cluster_handler(CLUSTER_HANDLER_LEVEL, 8)
    ch_3 = cluster_handler(CLUSTER_HANDLER_COLOR, 768)
    ch_3.async_configure = AsyncMock(side_effect=TimeoutError)
    ch_3.async_initialize = AsyncMock(side_effect=TimeoutError)
    ch_4 = cluster_handler(CLUSTER_HANDLER_ON_OFF, 6)
    ch_5 = cluster_handler(CLUSTER_HANDLER_LEVEL, 8)
    ch_5.async_configure = AsyncMock(side_effect=TimeoutError)
    ch_5.async_initialize = AsyncMock(side_effect=TimeoutError)

    endpoint_mock = mock.MagicMock(spec_set=ZigpyEndpoint)
    type(endpoint_mock).in_clusters = mock.PropertyMock(return_value={})
    type(endpoint_mock).out_clusters = mock.PropertyMock(return_value={})
    type(endpoint_mock).profile_id = mock.PropertyMock(
        return_value=zigpy.profiles.zha.PROFILE_ID
    )
    type(endpoint_mock).device_type = mock.PropertyMock(
        return_value=zigpy.profiles.zha.DeviceType.COLOR_DIMMABLE_LIGHT
    )
    zha_dev = mock.MagicMock(spec=Device)
    zha_dev.unique_id = "00:11:22:33:44:55:66:77"
    type(zha_dev).quirk_id = mock.PropertyMock(return_value=None)
    endpoint = Endpoint.new(endpoint_mock, zha_dev)

    claimed = {ch_1.id: ch_1, ch_2.id: ch_2, ch_3.id: ch_3}
    client_handlers = {ch_4.id: ch_4, ch_5.id: ch_5}

    with (
        mock.patch.dict(endpoint.claimed_cluster_handlers, claimed, clear=True),
        mock.patch.dict(endpoint.client_cluster_handlers, client_handlers, clear=True),
    ):
        await endpoint.async_configure()
        await endpoint.async_initialize(mock.sentinel.from_cache)

    for ch in [*claimed.values(), *client_handlers.values()]:
        assert ch.async_initialize.call_count == 1
        assert ch.async_initialize.await_count == 1
        assert ch.async_initialize.call_args[0][0] is mock.sentinel.from_cache
        assert ch.async_configure.call_count == 1
        assert ch.async_configure.await_count == 1

    assert ch_3.debug.call_count == 2
    assert ch_5.debug.call_count == 2


def zigpy_zll_device_mock(zha_gateway: Gateway) -> ZigpyDevice:
    """ZLL device fixture."""

    return create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [0x1000],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: 0x1234,
                SIG_EP_PROFILE: zigpy.profiles.zll.PROFILE_ID,
            }
        },
        "00:11:22:33:44:55:66:77",
        "test manufacturer",
        "test model",
    )


async def test_zll_device_groups(zha_gateway: Gateway) -> None:
    """Test adding coordinator to ZLL groups."""
    zigpy_zll_device: ZigpyDevice = zigpy_zll_device_mock(zha_gateway)
    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)

    cluster = zigpy_zll_device.endpoints[1].lightlink
    cluster_handler = LightLinkClusterHandler(cluster, endpoint)
    get_group_identifiers_rsp = zigpy.zcl.clusters.lightlink.LightLink.commands_by_name[
        "get_group_identifiers_rsp"
    ].schema

    with patch.object(
        cluster,
        "get_group_identifiers",
        AsyncMock(
            return_value=get_group_identifiers_rsp(
                total=0, start_index=0, group_info_records=[]
            )
        ),
    ) as get_group_identifiers:
        await cluster_handler.async_configure()
        assert get_group_identifiers.await_count == 1
        assert cluster.bind.call_count == 0
        assert zigpy_coordinator_device.add_to_group.await_count == 1
        assert zigpy_coordinator_device.add_to_group.await_args[0][0] == 0x0000

    zigpy_coordinator_device.add_to_group.reset_mock()
    group_1 = zigpy.zcl.clusters.lightlink.GroupInfoRecord(0xABCD, 0x00)
    group_2 = zigpy.zcl.clusters.lightlink.GroupInfoRecord(0xAABB, 0x00)
    with patch.object(
        cluster,
        "get_group_identifiers",
        AsyncMock(
            return_value=get_group_identifiers_rsp(
                total=2, start_index=0, group_info_records=[group_1, group_2]
            )
        ),
    ) as get_group_identifiers:
        await cluster_handler.async_configure()
        assert get_group_identifiers.await_count == 1
        assert cluster.bind.call_count == 0
        assert zigpy_coordinator_device.add_to_group.await_count == 2
        assert (
            zigpy_coordinator_device.add_to_group.await_args_list[0][0][0]
            == group_1.group_id
        )
        assert (
            zigpy_coordinator_device.add_to_group.await_args_list[1][0][0]
            == group_2.group_id
        )


@mock.patch(
    "zha.application.discovery.ENDPOINT_PROBE.discover_entities",
    mock.MagicMock(),
)
async def test_cluster_no_ep_attribute(
    zha_gateway: Gateway,  # pylint: disable=unused-argument
) -> None:
    """Test cluster handlers for clusters without ep_attribute."""

    zha_device = await join_zigpy_device(
        zha_gateway,
        create_mock_zigpy_device(
            zha_gateway,
            {
                1: {
                    SIG_EP_INPUT: [0x042E],
                    SIG_EP_OUTPUT: [],
                    SIG_EP_TYPE: 0x1234,
                    SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
                }
            },
        ),
    )

    assert "1:0x042e" in zha_device._endpoints[1].all_cluster_handlers
    assert zha_device._endpoints[1].all_cluster_handlers["1:0x042e"].name


async def test_configure_reporting(zha_gateway: Gateway) -> None:
    """Test setting up a cluster handler and configuring attribute reporting in two batches."""

    zigpy_coordinator_device: ZigpyDevice = zigpy_coordinator_device_mock(zha_gateway)
    endpoint: Endpoint = endpoint_mock(zigpy_coordinator_device)

    class TestZigbeeClusterHandler(ClusterHandler):
        """Test cluster handler that requests reporting for four attributes."""

        BIND = True
        REPORT_CONFIG = (
            # By name
            AttrReportConfig(attr="current_x", config=(1, 60, 1)),
            AttrReportConfig(attr="current_hue", config=(1, 60, 2)),
            AttrReportConfig(attr="color_temperature", config=(1, 60, 3)),
            AttrReportConfig(attr="current_y", config=(1, 60, 4)),
        )

    mock_ep = mock.AsyncMock()
    mock_ep.profile_id = zigpy.profiles.zha.PROFILE_ID
    mock_ep.device.zdo = AsyncMock()

    cluster = zigpy.zcl.clusters.lighting.Color(mock_ep)
    cluster.bind = AsyncMock(
        spec_set=cluster.bind,
        return_value=[zdo_t.Status.SUCCESS],  # ZDOCmd.Bind_rsp
    )
    cluster.configure_reporting_multiple = AsyncMock(
        spec_set=cluster.configure_reporting_multiple,
        return_value=[
            foundation.ConfigureReportingResponseRecord(
                status=foundation.Status.SUCCESS
            )
        ],
    )

    cluster_handler = TestZigbeeClusterHandler(cluster, endpoint)
    await cluster_handler.async_configure()

    # Since we request reporting for five attributes, we need to make two calls (3 + 1)
    assert cluster.configure_reporting_multiple.mock_calls == [
        mock.call(
            {
                "current_x": (1, 60, 1),
                "current_hue": (1, 60, 2),
                "color_temperature": (1, 60, 3),
            }
        ),
        mock.call(
            {
                "current_y": (1, 60, 4),
            }
        ),
    ]


async def test_invalid_cluster_handler(zha_gateway: Gateway, caplog) -> None:  # pylint: disable=unused-argument
    """Test setting up a cluster handler that fails to match properly."""

    class TestZigbeeClusterHandler(ClusterHandler):
        """Test cluster handler that fails to match properly."""

        REPORT_CONFIG = (AttrReportConfig(attr="missing_attr", config=(1, 60, 1)),)

    mock_device = mock.AsyncMock(spec_set=zigpy.device.Device)
    zigpy_ep = zigpy.endpoint.Endpoint(mock_device, endpoint_id=1)
    zigpy_ep.profile_id = zigpy.profiles.zha.PROFILE_ID

    cluster = zigpy_ep.add_input_cluster(zigpy.zcl.clusters.lighting.Color.cluster_id)
    cluster.configure_reporting_multiple = AsyncMock(
        spec_set=cluster.configure_reporting_multiple,
        return_value=[
            foundation.ConfigureReportingResponseRecord(
                status=foundation.Status.SUCCESS
            )
        ],
    )

    mock_zha_device = mock.AsyncMock(spec=Device)
    mock_zha_device.quirk_id = None
    mock_zha_device.unique_id = "aa:bb:cc:dd:11:22:33:44"

    zha_endpoint = Endpoint(zigpy_ep, mock_zha_device)

    # The cluster handler throws an error when matching this cluster
    with pytest.raises(KeyError):
        TestZigbeeClusterHandler(cluster, zha_endpoint)

    # And one is also logged at runtime
    with (
        patch.dict(
            CLUSTER_HANDLER_REGISTRY[cluster.cluster_id],
            {None: TestZigbeeClusterHandler},
        ),
        caplog.at_level(logging.WARNING),
    ):
        zha_endpoint.add_all_cluster_handlers()

    assert "missing_attr" in caplog.text


async def test_standard_cluster_handler(
    zha_gateway: Gateway,
) -> None:  # pylint: disable=unused-argument
    """Test setting up a cluster handler that matches a standard cluster."""

    class TestZigbeeClusterHandler(ColorClusterHandler):
        """Test cluster handler that matches a standard cluster."""

    mock_device = mock.AsyncMock(spec_set=zigpy.device.Device)
    zigpy_ep = zigpy.endpoint.Endpoint(mock_device, endpoint_id=1)
    zigpy_ep.profile_id = zigpy.profiles.zha.PROFILE_ID

    cluster = zigpy_ep.add_input_cluster(zigpy.zcl.clusters.lighting.Color.cluster_id)
    cluster.configure_reporting_multiple = AsyncMock(
        spec_set=cluster.configure_reporting_multiple,
        return_value=[
            foundation.ConfigureReportingResponseRecord(
                status=foundation.Status.SUCCESS
            )
        ],
    )

    mock_zha_device = mock.AsyncMock(spec=Device)
    mock_zha_device.quirk_id = None
    mock_zha_device.unique_id = "aa:bb:cc:dd:11:22:33:44"

    zha_endpoint = Endpoint(zigpy_ep, mock_zha_device)

    with patch.dict(
        CLUSTER_HANDLER_REGISTRY[cluster.cluster_id],
        {"__test_quirk_id": TestZigbeeClusterHandler},
    ):
        zha_endpoint.add_all_cluster_handlers()

    assert len(zha_endpoint.all_cluster_handlers) == 1
    assert isinstance(
        list(zha_endpoint.all_cluster_handlers.values())[0], ColorClusterHandler
    )


async def test_quirk_id_cluster_handler(
    zha_gateway: Gateway,
) -> None:  # pylint: disable=unused-argument
    """Test setting up a cluster handler that matches a standard cluster."""

    class TestZigbeeClusterHandler(ColorClusterHandler):
        """Test cluster handler that matches a standard cluster."""

    mock_device = mock.AsyncMock(spec_set=zigpy.device.Device)
    zigpy_ep = zigpy.endpoint.Endpoint(mock_device, endpoint_id=1)
    zigpy_ep.profile_id = zigpy.profiles.zha.PROFILE_ID

    cluster = zigpy_ep.add_input_cluster(zigpy.zcl.clusters.lighting.Color.cluster_id)
    cluster.configure_reporting_multiple = AsyncMock(
        spec_set=cluster.configure_reporting_multiple,
        return_value=[
            foundation.ConfigureReportingResponseRecord(
                status=foundation.Status.SUCCESS
            )
        ],
    )

    mock_zha_device = mock.AsyncMock(spec=Device)
    mock_zha_device.unique_id = "aa:bb:cc:dd:11:22:33:44"
    mock_zha_device.quirk_id = "__test_quirk_id"
    zha_endpoint = Endpoint(zigpy_ep, mock_zha_device)

    with patch.dict(
        CLUSTER_HANDLER_REGISTRY[cluster.cluster_id],
        {"__test_quirk_id": TestZigbeeClusterHandler},
    ):
        zha_endpoint.add_all_cluster_handlers()

    assert len(zha_endpoint.all_cluster_handlers) == 1
    assert isinstance(
        list(zha_endpoint.all_cluster_handlers.values())[0], TestZigbeeClusterHandler
    )


# parametrize side effects:
@pytest.mark.parametrize(
    ("side_effect", "expected_error"),
    [
        (zigpy.exceptions.ZigbeeException(), "Failed to send request"),
        (
            zigpy.exceptions.ZigbeeException("Zigbee exception"),
            "Failed to send request: Zigbee exception",
        ),
        (TimeoutError(), "Failed to send request: device did not respond"),
    ],
)
async def test_retry_request(
    side_effect: Exception, expected_error: str | None
) -> None:
    """Test the `retry_request` decorator's handling of zigpy-internal exceptions."""

    async def func(arg1: int, arg2: int) -> int:
        assert arg1 == 1
        assert arg2 == 2

        raise side_effect

    func = mock.AsyncMock(wraps=func)
    decorated_func = retry_request(func)

    with pytest.raises(ZHAException) as exc:
        await decorated_func(1, arg2=2)

    assert func.await_count == 3
    assert isinstance(exc.value, ZHAException)
    assert str(exc.value) == expected_error


async def test_cluster_handler_naming() -> None:
    """Test that all cluster handlers are named appropriately."""
    for client_cluster_handler in CLIENT_CLUSTER_HANDLER_REGISTRY.values():
        assert issubclass(client_cluster_handler, ClientClusterHandler)
        assert client_cluster_handler.__name__.endswith("ClientClusterHandler")

    for cluster_handler_dict in CLUSTER_HANDLER_REGISTRY.values():
        for cluster_handler in cluster_handler_dict.values():
            assert not issubclass(cluster_handler, ClientClusterHandler)
            assert cluster_handler.__name__.endswith("ClusterHandler")


async def test_parse_and_log_command(zha_gateway: Gateway) -> None:
    """Test that `parse_and_log_command` correctly parses a known command."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/ikea-of-sweden-rodret-dimmer.json",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    identify_entity = get_entity(
        zha_device, platform=Platform.BUTTON, entity_type=IdentifyButton
    )
    identify_handler = identify_entity._cluster_handler

    assert parse_and_log_command(identify_handler, 0x00, 0x01, []) == "identify_query"


async def test_parse_and_log_command_unknown(zha_gateway: Gateway):
    """Test that `parse_and_log_command` correctly parses an unknown command."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/ikea-of-sweden-rodret-dimmer.json",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    identify_entity = get_entity(
        zha_device, platform=Platform.BUTTON, entity_type=IdentifyButton
    )
    identify_handler = identify_entity._cluster_handler

    assert parse_and_log_command(identify_handler, 0x00, 0xAB, []) == "0xAB"


async def test_zha_send_event_from_quirk(zha_gateway: Gateway):
    """Test that a quirk can send an event."""
    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    Basic.cluster_id,
                    PowerConfiguration.cluster_id,
                    OnOff.cluster_id,
                    MultistateInput.cluster_id,
                ],
                SIG_EP_OUTPUT: [Basic.cluster_id],
                SIG_EP_TYPE: BUTTON_DEVICE_TYPE,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        model="lumi.sensor_switch.aq3",
        manufacturer="LUMI",
        quirk=SwitchAQ3,
    )

    assert isinstance(zigpy_device, SwitchAQ3)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    ms_input_ch = zha_device.endpoints[1].all_cluster_handlers["1:0x0012"]
    assert ms_input_ch is not None

    ms_input_ch.zha_send_event = MagicMock(wraps=ms_input_ch.zha_send_event)

    await send_attributes_report(zha_gateway, ms_input_ch.cluster, {0x0055: 0x01})

    assert ms_input_ch.zha_send_event.call_count == 1
    assert ms_input_ch.zha_send_event.mock_calls == [
        call(
            "single",
            {
                "value": 1.0,
            },
        )
    ]

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    Basic.cluster_id,
                    PowerConfiguration.cluster_id,
                    Identify.cluster_id,
                    PollControl.cluster_id,
                    TemperatureMeasurement.cluster_id,
                    Diagnostic.cluster_id,
                ],
                SIG_EP_OUTPUT: [
                    Identify.cluster_id,
                    OnOff.cluster_id,
                    LevelControl.cluster_id,
                    Ota.cluster_id,
                ],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.LEVEL_CONTROL_SWITCH,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        model="LIGHTIFY Dimming Switch",
        manufacturer="OSRAM",
        quirk=CentraLite3130,
        ieee="00:15:8e:01:01:02:03:04",
    )

    assert isinstance(zigpy_device, CentraLite3130)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    on_off_ch = zha_device.endpoints[1].client_cluster_handlers["1:0x0006_client"]
    assert on_off_ch is not None

    on_off_ch.emit_zha_event = MagicMock(wraps=on_off_ch.emit_zha_event)
    on_off_ch.emit_zha_event.reset_mock()

    on_off_ch.cluster_command(1, OnOff.ServerCommandDefs.on.id, [])

    assert on_off_ch.emit_zha_event.call_count == 1
    assert on_off_ch.emit_zha_event.mock_calls == [call("on", [])]
    on_off_ch.emit_zha_event.reset_mock()

    await send_attributes_report(
        zha_gateway, on_off_ch.cluster, {OnOff.AttributeDefs.on_off.name: 0x01}
    )

    assert on_off_ch.emit_zha_event.call_count == 1
    assert on_off_ch.emit_zha_event.mock_calls == [
        call(
            "attribute_updated",
            {
                "attribute_id": 0,
                "attribute_name": "on_off",
                "attribute_value": True,
                "value": True,
            },
        )
    ]

    on_off_ch.emit_zha_event.reset_mock()

    await send_attributes_report(zha_gateway, on_off_ch.cluster, {0x25: "Bar"})

    assert on_off_ch.emit_zha_event.call_count == 1
    assert on_off_ch.emit_zha_event.mock_calls == [
        call(
            "attribute_updated",
            {
                "attribute_id": 0x25,
                "attribute_name": "Unknown",
                "attribute_value": "Bar",
                "value": "Bar",
            },
        )
    ]


async def test_zdo_cluster_handler(zha_gateway: Gateway):
    """Test that a quirk can send an event."""
    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    Basic.cluster_id,
                    PowerConfiguration.cluster_id,
                    OnOff.cluster_id,
                    MultistateInput.cluster_id,
                ],
                SIG_EP_OUTPUT: [Basic.cluster_id],
                SIG_EP_TYPE: BUTTON_DEVICE_TYPE,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        model="lumi.sensor_switch.aq3",
        manufacturer="LUMI",
        quirk=SwitchAQ3,
    )

    assert isinstance(zigpy_device, SwitchAQ3)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    assert zha_device.zdo_cluster_handler is not None
    assert zha_device.zdo_cluster_handler.status == ClusterHandlerStatus.INITIALIZED
    assert zha_device.zdo_cluster_handler.cluster is not None
    assert zha_device.zdo_cluster_handler.cluster == zigpy_device.endpoints[0]
    assert (
        zha_device.zdo_cluster_handler.unique_id
        == f"{str(zha_device.ieee)}:{zha_device.name}_ZDO"
    )



"""Regenerate device diagnostics JSON, when new entities are added."""

import sys

sys.path.insert(0, "tests")

import asyncio
import contextlib
import json
import logging
import pathlib
from unittest.mock import patch

from tests.common import ZhaJsonEncoder, join_zigpy_device, zigpy_device_from_json
from tests.conftest import TestGateway, make_zha_data, make_zigpy_app_controller

REPO_ROOT = pathlib.Path(__file__).parent.parent
_LOGGER = logging.getLogger(__name__)


@contextlib.asynccontextmanager
async def create_zha_gateway():
    """Turn a pytest fixture into a normal context manager."""
    # This isn't the way Pytest is meant to be used :)
    with make_zigpy_app_controller() as zigpy_app_controller:
        async with TestGateway(
            data=make_zha_data(),
            app=zigpy_app_controller,
        ) as zha_gateway:
            yield zha_gateway


async def main(files: list[str] | None = None) -> None:
    """Entry point."""
    if files is None:
        paths = list((REPO_ROOT / "tests" / "data" / "devices").glob("**/*.json"))
    else:
        paths = [pathlib.Path(f) for f in files]

    async with create_zha_gateway() as zha_gateway:
        for device_json in paths:
            _LOGGER.info("Migrating %s", device_json)
            zigpy_device = await zigpy_device_from_json(
                zha_gateway.application_controller,
                device_json,
            )

            with patch("zigpy.zcl.Cluster._update_attribute"):
                zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

            new_json = json.dumps(
                zha_device.get_diagnostics_json(), indent=2, cls=ZhaJsonEncoder
            )
            device_json.write_text(new_json)

            await zha_device.on_remove()


if __name__ == "__main__":
    import coloredlogs

    coloredlogs.install(level="DEBUG")

    asyncio.run(
        main(
            files=(sys.argv[1:] if len(sys.argv) > 1 else None),
        )
    )



"""Device for Zigbee Home Automation."""

# pylint: disable=too-many-lines

from __future__ import annotations

import asyncio
from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator
import copy
import dataclasses
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
import logging
import time
from typing import TYPE_CHECKING, Any, Final, Self

from zigpy.device import Device as ZigpyDevice
import zigpy.exceptions
from zigpy.profiles import PROFILES
import zigpy.quirks
from zigpy.quirks.v2 import CustomDeviceV2, DeviceAlertMetadata, QuirksV2RegistryEntry
from zigpy.types import uint1_t, uint8_t, uint16_t
from zigpy.types.named import EUI64, NWK, ExtendedPanId
from zigpy.zcl.clusters import Cluster
from zigpy.zcl.clusters.general import Groups, Identify
from zigpy.zcl.foundation import (
    Status as ZclStatus,
    WriteAttributesResponse,
    ZCLCommandDef,
)
import zigpy.zdo.types as zdo_types
from zigpy.zdo.types import RouteStatus, _NeighborEnums

from zha.application import Platform, discovery
from zha.application.const import (
    ATTR_ARGS,
    ATTR_ATTRIBUTE,
    ATTR_CLUSTER_ID,
    ATTR_CLUSTER_TYPE,
    ATTR_COMMAND,
    ATTR_COMMAND_TYPE,
    ATTR_ENDPOINT_ID,
    ATTR_ENDPOINTS,
    ATTR_MANUFACTURER,
    ATTR_MODEL,
    ATTR_NODE_DESCRIPTOR,
    ATTR_PARAMS,
    ATTR_QUIRK_ID,
    ATTR_VALUE,
    CLUSTER_COMMAND_SERVER,
    CLUSTER_COMMANDS_CLIENT,
    CLUSTER_COMMANDS_SERVER,
    CLUSTER_TYPE_IN,
    CLUSTER_TYPE_OUT,
    POWER_BATTERY_OR_UNKNOWN,
    POWER_MAINS_POWERED,
    UNKNOWN,
    UNKNOWN_MANUFACTURER,
    UNKNOWN_MODEL,
    ZHA_CLUSTER_HANDLER_CFG_DONE,
    ZHA_CLUSTER_HANDLER_MSG,
    ZHA_DEVICE_UPDATED_EVENT,
    ZHA_EVENT,
)
from zha.application.helpers import convert_to_zcl_values, convert_zcl_value
from zha.application.platforms import (
    BaseEntity,
    BaseEntityInfo,
    EntityStateChangedEvent,
    PlatformEntity,
)
from zha.application.platforms.update import BaseFirmwareUpdateEntity
from zha.const import STATE_CHANGED
from zha.event import EventBase
from zha.exceptions import ZHAException
from zha.mixins import LogMixin
from zha.zigbee.cluster_handlers import ClusterHandler, ZDOClusterHandler
from zha.zigbee.endpoint import Endpoint

if TYPE_CHECKING:
    from zha.application.gateway import Gateway

_LOGGER = logging.getLogger(__name__)
_CHECKIN_GRACE_PERIODS = 2
DIAGNOSTICS_JSON_VERSION = 1


def get_cluster_attr_data(cluster: Cluster) -> list[dict]:
    """Return cluster attribute data."""
    attributes_info = []

    for attr_def in cluster.attributes.values():
        info = {
            "id": f"0x{attr_def.id:04x}",
            "name": attr_def.name,
            "zcl_type": (
                attr_def.zcl_type.name if attr_def.zcl_type.name != "bool_" else "bool"
            ),
            "value": cluster.get(attr_def.name),
            "unsupported": (attr_def.id in cluster.unsupported_attributes),
        }

        # Don't unnecessarily list out attributes that are just unread
        if info["value"] is None and not info["unsupported"]:
            continue

        # Delete unused keys
        if info["value"] is not None:
            del info["unsupported"]
        else:
            del info["value"]

        attributes_info.append(info)

    return attributes_info


def get_device_automation_triggers(
    device: zigpy.device.Device,
) -> dict[tuple[str, str], dict[str, str]]:
    """Get the supported device automation triggers for a zigpy device."""
    return {
        ("device_offline", "device_offline"): {"device_event_type": "device_offline"},
        **getattr(device, "device_automation_triggers", {}),
    }


@dataclass(frozen=True, kw_only=True)
class ClusterBinding:
    """Describes a cluster binding."""

    name: str
    type: str
    id: int
    endpoint_id: int


class DeviceStatus(Enum):
    """Status of a device."""

    CREATED = 1
    INITIALIZED = 2


@dataclass(kw_only=True, frozen=True)
class ZHAEvent:
    """Event generated when a device wishes to send an arbitrary event."""

    device_ieee: EUI64
    unique_id: str
    data: dict[str, Any]
    event_type: Final[str] = ZHA_EVENT
    event: Final[str] = ZHA_EVENT


@dataclass(kw_only=True, frozen=True)
class DeviceFirmwareInfoUpdatedEvent:
    """Event generated when the device firmware information has changed."""

    event_type: Final[str] = ZHA_DEVICE_UPDATED_EVENT
    event: Final[str] = ZHA_DEVICE_UPDATED_EVENT

    old_firmware_version: str | None
    new_firmware_version: str | None


@dataclass(kw_only=True, frozen=True)
class ClusterHandlerConfigurationComplete:
    """Event generated when all cluster handlers are configured."""

    device_ieee: EUI64
    unique_id: str
    event_type: Final[str] = ZHA_CLUSTER_HANDLER_MSG
    event: Final[str] = ZHA_CLUSTER_HANDLER_CFG_DONE


@dataclass(kw_only=True, frozen=True)
class DeviceInfo:
    """Describes a device."""

    ieee: EUI64
    nwk: NWK
    manufacturer: str
    model: str
    name: str
    quirk_applied: bool
    quirk_class: str
    quirk_id: str | None
    manufacturer_code: int | None
    power_source: str
    lqi: int
    rssi: int
    last_seen: str
    available: bool
    device_type: str
    signature: dict[str, Any]


@dataclass(kw_only=True, frozen=True)
class NeighborInfo:
    """Describes a neighbor."""

    device_type: _NeighborEnums.DeviceType
    rx_on_when_idle: _NeighborEnums.RxOnWhenIdle
    relationship: _NeighborEnums.Relationship
    extended_pan_id: ExtendedPanId
    ieee: EUI64
    nwk: NWK
    permit_joining: _NeighborEnums.PermitJoins
    depth: uint8_t
    lqi: uint8_t


@dataclass(kw_only=True, frozen=True)
class RouteInfo:
    """Describes a route."""

    dest_nwk: NWK
    route_status: RouteStatus
    memory_constrained: uint1_t
    many_to_one: uint1_t
    route_record_required: uint1_t
    next_hop: NWK


@dataclass(kw_only=True, frozen=True)
class EndpointNameInfo:
    """Describes an endpoint name."""

    name: str


@dataclass(kw_only=True, frozen=True)
class ExtendedDeviceInfo(DeviceInfo):
    """Describes a ZHA device."""

    active_coordinator: bool
    entities: dict[str, BaseEntityInfo]
    neighbors: list[NeighborInfo]
    routes: list[RouteInfo]
    endpoint_names: list[EndpointNameInfo]


class Device(LogMixin, EventBase):
    """ZHA Zigbee device object."""

    unique_id: str

    def __init__(
        self,
        zigpy_device: zigpy.device.Device,
        _gateway: Gateway,
    ) -> None:
        """Initialize the gateway."""
        super().__init__()

        self.unique_id = str(zigpy_device.ieee)

        self._gateway: Gateway = _gateway
        self._zigpy_device: ZigpyDevice = zigpy_device
        self.quirk_applied: bool = isinstance(
            self._zigpy_device, zigpy.quirks.BaseCustomDevice
        )
        self.quirk_class: str = (
            f"{self._zigpy_device.__class__.__module__}."
            f"{self._zigpy_device.__class__.__name__}"
        )
        self.quirk_id: str | None = getattr(self._zigpy_device, ATTR_QUIRK_ID, None)
        self._power_config_ch: ClusterHandler | None = None
        self._identify_ch: ClusterHandler | None = None
        self._basic_ch: ClusterHandler | None = None
        self._firmware_version: str | None = None

        device_options = _gateway.config.config.device_options
        if self.is_mains_powered:
            self.consider_unavailable_time: int = (
                device_options.consider_unavailable_mains
            )
        else:
            self.consider_unavailable_time = device_options.consider_unavailable_battery
        self._available: bool = self.is_active_coordinator or (
            self.last_seen is not None
            and time.time() - self.last_seen < self.consider_unavailable_time
        )
        self._checkins_missed_count: int = 0
        self._on_network: bool = True

        self._platform_entities: dict[tuple[Platform, str], PlatformEntity] = {}
        self._pending_entities: list[PlatformEntity] = []
        self.semaphore: asyncio.Semaphore = asyncio.Semaphore(3)

        self._on_remove_callbacks: list[Callable[[], None]] = []

        self._zdo_handler: ZDOClusterHandler = ZDOClusterHandler(self)
        self._zdo_handler.on_add()
        self._on_remove_callbacks.append(self._zdo_handler.on_remove)

        self.status: DeviceStatus = DeviceStatus.CREATED

        self._endpoints: dict[int, Endpoint] = {}
        for ep_id, endpoint in zigpy_device.endpoints.items():
            if ep_id != 0:
                ep = Endpoint.new(endpoint, self)
                self._endpoints[ep_id] = ep
                self._on_remove_callbacks.append(ep.on_remove)

    def __repr__(self) -> str:
        """Return a string representation of the device."""
        return (
            f"{repr(self._zigpy_device)} - "
            f"quirk_applied: {self.quirk_applied} - "
            f"quirk_or_device_class: {self.quirk_class} - "
            f"quirk_id: {self.quirk_id}"
        )

    @property
    def device(self) -> zigpy.device.Device:
        """Return underlying Zigpy device."""
        return self._zigpy_device

    @cached_property
    def name(self) -> str:
        """Return device name."""
        return f"{self.manufacturer} {self.model}"

    @property
    def ieee(self) -> EUI64:
        """Return ieee address for device."""
        return self._zigpy_device.ieee

    @property
    def quirk_metadata(self) -> QuirksV2RegistryEntry | None:
        """Return the quirk metadata for this device."""
        return getattr(self._zigpy_device, "quirk_metadata", None)

    @cached_property
    def manufacturer(self) -> str:
        """Return manufacturer for device."""
        if self.is_active_coordinator:
            manufacturer = (
                self.gateway.application_controller.state.node_info.manufacturer
            )
            if manufacturer is None:
                return ""
            return manufacturer

        if (
            self.quirk_metadata is not None
            and self.quirk_metadata.friendly_name is not None
        ):
            return self.quirk_metadata.friendly_name.manufacturer

        if self._zigpy_device.manufacturer is None:
            return UNKNOWN_MANUFACTURER

        return self._zigpy_device.manufacturer

    @cached_property
    def model(self) -> str:
        """Return model for device."""
        if self.is_active_coordinator:
            model = self.gateway.application_controller.state.node_info.model
            if model is None:
                return f"Generic Zigbee Coordinator ({self.gateway.radio_type.pretty_name})"
            return model

        if (
            self.quirk_metadata is not None
            and self.quirk_metadata.friendly_name is not None
        ):
            return self.quirk_metadata.friendly_name.model

        if self._zigpy_device.model is None:
            return UNKNOWN_MODEL

        return self._zigpy_device.model

    @cached_property
    def device_alerts(self) -> Iterable[DeviceAlertMetadata]:
        """Return device alerts for this device."""
        if self.quirk_metadata is None:
            return []

        return self.quirk_metadata.device_alerts

    @cached_property
    def manufacturer_code(self) -> int | None:
        """Return the manufacturer code for the device."""
        if self._zigpy_device.node_desc is None:
            return None

        return self._zigpy_device.node_desc.manufacturer_code

    @property
    def nwk(self) -> NWK:
        """Return nwk for device."""
        return self._zigpy_device.nwk

    @property
    def lqi(self):
        """Return lqi for device."""
        return self._zigpy_device.lqi

    @property
    def rssi(self):
        """Return rssi for device."""
        return self._zigpy_device.rssi

    @property
    def last_seen(self) -> float | None:
        """Return last_seen for device."""
        return self._zigpy_device.last_seen

    @cached_property
    def is_mains_powered(self) -> bool | None:
        """Return true if device is mains powered."""
        if self._zigpy_device.node_desc is None:
            return None

        return self._zigpy_device.node_desc.is_mains_powered

    @cached_property
    def device_type(self) -> str:
        """Return the logical device type for the device."""
        if self._zigpy_device.node_desc is None:
            return UNKNOWN

        return self._zigpy_device.node_desc.logical_type.name

    @property
    def power_source(self) -> str:
        """Return the power source for the device."""
        return (
            POWER_MAINS_POWERED if self.is_mains_powered else POWER_BATTERY_OR_UNKNOWN
        )

    @cached_property
    def is_router(self) -> bool | None:
        """Return true if this is a routing capable device."""
        if self._zigpy_device.node_desc is None:
            return None

        return self._zigpy_device.node_desc.is_router

    @cached_property
    def is_coordinator(self) -> bool | None:
        """Return true if this device represents a coordinator."""
        if self._zigpy_device.node_desc is None:
            return None

        return self._zigpy_device.node_desc.is_coordinator

    @property
    def is_active_coordinator(self) -> bool:
        """Return true if this device is the active coordinator."""
        if not self.is_coordinator:
            return False

        return self.ieee == self.gateway.state.node_info.ieee

    @cached_property
    def is_end_device(self) -> bool | None:
        """Return true if this device is an end device."""
        if self._zigpy_device.node_desc is None:
            return None

        return self._zigpy_device.node_desc.is_end_device

    @property
    def is_groupable(self) -> bool:
        """Return true if this device has a group cluster."""
        return self.is_active_coordinator or (
            self.available and bool(self.async_get_groupable_endpoints())
        )

    @cached_property
    def skip_configuration(self) -> bool:
        """Return true if the device should not issue configuration related commands."""
        return self._zigpy_device.skip_configuration or bool(self.is_active_coordinator)

    @property
    def gateway(self):
        """Return the gateway for this device."""
        return self._gateway

    @cached_property
    def device_automation_commands(self) -> dict[str, list[tuple[str, str]]]:
        """Return the a lookup of commands to etype/sub_type."""
        commands: dict[str, list[tuple[str, str]]] = {}
        for etype_subtype, trigger in self.device_automation_triggers.items():
            if command := trigger.get(ATTR_COMMAND):
                commands.setdefault(command, []).append(etype_subtype)
        return commands

    @cached_property
    def device_automation_triggers(self) -> dict[tuple[str, str], dict[str, str]]:
        """Return the device automation triggers for this device."""
        return get_device_automation_triggers(self._zigpy_device)

    @property
    def available(self):
        """Return True if device is available."""
        return self.is_active_coordinator or (self._available and self.on_network)

    @available.setter
    def available(self, new_availability: bool) -> None:
        """Set device availability."""
        self._available = new_availability

    @property
    def on_network(self):
        """Return True if device is currently on the network."""
        return self.is_active_coordinator or self._on_network

    @on_network.setter
    def on_network(self, new_on_network: bool) -> None:
        """Set device on_network flag."""
        self.update_available(new_on_network)
        self._on_network = new_on_network
        if not new_on_network:
            self.debug("Device is not on the network, marking unavailable")

    @property
    def power_configuration_ch(self) -> ClusterHandler | None:
        """Return power configuration cluster handler."""
        return self._power_config_ch

    @power_configuration_ch.setter
    def power_configuration_ch(self, cluster_handler: ClusterHandler) -> None:
        """Power configuration cluster handler setter."""
        if self._power_config_ch is None:
            self._power_config_ch = cluster_handler

    @property
    def basic_ch(self) -> ClusterHandler | None:
        """Return basic cluster handler."""
        return self._basic_ch

    @basic_ch.setter
    def basic_ch(self, cluster_handler: ClusterHandler) -> None:
        """Set the basic cluster handler."""
        if self._basic_ch is None:
            self._basic_ch = cluster_handler

    @property
    def identify_ch(self) -> ClusterHandler | None:
        """Return power configuration cluster handler."""
        return self._identify_ch

    @identify_ch.setter
    def identify_ch(self, cluster_handler: ClusterHandler) -> None:
        """Power configuration cluster handler setter."""
        if self._identify_ch is None:
            self._identify_ch = cluster_handler

    @property
    def zdo_cluster_handler(self) -> ZDOClusterHandler:
        """Return ZDO cluster handler."""
        return self._zdo_handler

    @property
    def endpoints(self) -> dict[int, Endpoint]:
        """Return the endpoints for this device."""
        return self._endpoints

    @cached_property
    def zigbee_signature(self) -> dict[str, Any]:
        """Get zigbee signature for this device."""
        return {
            ATTR_NODE_DESCRIPTOR: self._zigpy_device.node_desc,
            ATTR_ENDPOINTS: {
                signature[0]: signature[1]
                for signature in [
                    endpoint.zigbee_signature for endpoint in self._endpoints.values()
                ]
            },
            ATTR_MANUFACTURER: self.manufacturer,
            ATTR_MODEL: self.model,
        }

    @property
    def firmware_version(self) -> str | None:
        """Return the software version for this device."""
        return self._firmware_version

    @property
    def platform_entities(self) -> dict[tuple[Platform, str], PlatformEntity]:
        """Return the platform entities for this device."""
        return self._platform_entities

    def get_platform_entity(self, platform: Platform, unique_id: str) -> PlatformEntity:
        """Get a platform entity by unique id."""
        entity = self._platform_entities.get((platform, unique_id))
        if entity is None:
            raise KeyError(f"Entity {unique_id} not found")
        return entity

    @classmethod
    def new(
        cls,
        zigpy_dev: zigpy.device.Device,
        gateway: Gateway,
    ) -> Self:
        """Create new device."""
        return cls(zigpy_dev, gateway)

    def async_update_firmware_version(self, firmware_version: str) -> None:
        """Update device firmware version."""
        if firmware_version == self._firmware_version:
            return

        old_firmware_version = self._firmware_version
        self._firmware_version = firmware_version

        self.emit(
            DeviceFirmwareInfoUpdatedEvent.event_type,
            DeviceFirmwareInfoUpdatedEvent(
                old_firmware_version=old_firmware_version,
                new_firmware_version=firmware_version,
            ),
        )

    async def _check_available(self, *_: Any) -> None:
        # don't flip the availability state of the coordinator
        if self.is_active_coordinator:
            return
        if self.last_seen is None:
            self.debug("last_seen is None, marking the device unavailable")
            self.update_available(False)
            return

        difference = time.time() - self.last_seen
        if difference < self.consider_unavailable_time:
            self.debug(
                "Device seen - marking the device available and resetting counter"
            )
            self.update_available(True)
            self._checkins_missed_count = 0
            return

        if self._gateway.config.allow_polling:
            if (
                self._checkins_missed_count >= _CHECKIN_GRACE_PERIODS
                or self.manufacturer == "LUMI"
                or not self._endpoints
            ):
                self.debug(
                    (
                        "last_seen is %s seconds ago and ping attempts have been exhausted,"
                        " marking the device unavailable"
                    ),
                    difference,
                )
                self.update_available(False)
                return

            self._checkins_missed_count += 1
            self.debug(
                "Attempting to checkin with device - missed checkins: %s",
                self._checkins_missed_count,
            )
            if not self.basic_ch:
                self.debug("does not have a mandatory basic cluster")
                self.update_available(False)
                return
            res = await self.basic_ch.get_attribute_value(
                ATTR_MANUFACTURER, from_cache=False
            )
            if res is not None:
                self._checkins_missed_count = 0

    def update_available(self, available: bool) -> None:
        """Update device availability and signal entities."""
        self.debug(
            (
                "Update device availability -  device available: %s - new availability:"
                " %s - changed: %s"
            ),
            self.available,
            available,
            self.available ^ available,
        )
        availability_changed = self.available ^ available
        self.available = available
        if availability_changed and available:
            # reinit cluster handlers then signal entities
            self.debug(
                "Device availability changed and device became available,"
                " reinitializing cluster handlers"
            )
            self._gateway.async_create_task(
                self._async_became_available(),
                name=f"({self.nwk},{self.model})_async_became_available",
                eager_start=True,
            )
            return
        if availability_changed and not available:
            self.debug("Device availability changed and device became unavailable")
            for entity in self.platform_entities.values():
                entity.maybe_emit_state_changed_event()
            self.emit_zha_event(
                {
                    "device_event_type": "device_offline",
                },
            )

    def emit_zha_event(self, event_data: dict[str, str | int]) -> None:  # pylint: disable=unused-argument
        """Relay events directly."""
        self.emit(
            ZHA_EVENT,
            ZHAEvent(
                device_ieee=self.ieee,
                unique_id=str(self.ieee),
                data=event_data,
            ),
        )

    async def _async_became_available(self) -> None:
        """Update device availability and signal entities."""
        await self.async_initialize(False)
        for platform_entity in self._platform_entities.values():
            platform_entity.maybe_emit_state_changed_event()

    @property
    def device_info(self) -> DeviceInfo:
        """Return a device description for device."""
        ieee = self.ieee
        time_struct = time.localtime(self.last_seen)
        update_time = time.strftime("%Y-%m-%dT%H:%M:%S", time_struct)
        return DeviceInfo(
            ieee=ieee,
            nwk=self.nwk,
            manufacturer=self.manufacturer,
            model=self.model,
            name=self.name,
            quirk_applied=self.quirk_applied,
            quirk_class=self.quirk_class,
            quirk_id=self.quirk_id,
            manufacturer_code=self.manufacturer_code,
            power_source=self.power_source,
            lqi=self.lqi,
            rssi=self.rssi,
            last_seen=update_time,
            available=self.available,
            device_type=self.device_type,
            signature=self.zigbee_signature,
        )

    @property
    def extended_device_info(self) -> ExtendedDeviceInfo:
        """Get extended device information."""
        topology = self.gateway.application_controller.topology
        names: list[EndpointNameInfo] = []
        for endpoint in (ep for epid, ep in self.device.endpoints.items() if epid):
            profile = PROFILES.get(endpoint.profile_id)
            if profile and endpoint.device_type is not None:
                # DeviceType provides undefined enums
                names.append(
                    EndpointNameInfo(name=profile.DeviceType(endpoint.device_type).name)
                )
            else:
                names.append(
                    EndpointNameInfo(
                        name=(
                            f"unknown {endpoint.device_type} device_type "
                            f"of 0x{(endpoint.profile_id or 0xFFFF):04x} profile id"
                        )
                    )
                )

        return ExtendedDeviceInfo(
            **self.device_info.__dict__,
            active_coordinator=self.is_active_coordinator,
            entities={
                platform_entity.unique_id: platform_entity.info_object
                for platform_entity in self.platform_entities.values()
            },
            neighbors=[
                NeighborInfo(
                    device_type=neighbor.device_type.name,
                    rx_on_when_idle=neighbor.rx_on_when_idle.name,
                    relationship=neighbor.relationship.name,
                    extended_pan_id=neighbor.extended_pan_id,
                    ieee=neighbor.ieee,
                    nwk=neighbor.nwk,
                    permit_joining=neighbor.permit_joining.name,
                    depth=neighbor.depth,
                    lqi=neighbor.lqi,
                )
                for neighbor in topology.neighbors[self.ieee]
            ],
            routes=[
                RouteInfo(
                    dest_nwk=route.DstNWK,
                    route_status=route.RouteStatus.name,
                    memory_constrained=route.MemoryConstrained,
                    many_to_one=route.ManyToOne,
                    route_record_required=route.RouteRecordRequired,
                    next_hop=route.NextHop,
                )
                for route in topology.routes[self.ieee]
            ],
            endpoint_names=names,
        )

    async def async_configure(self) -> None:
        """Configure the device."""
        self.debug("started configuration")
        await self._zdo_handler.async_configure()
        self._zdo_handler.debug("'async_configure' stage succeeded")

        if isinstance(self._zigpy_device, zigpy.quirks.BaseCustomDevice):
            self.debug("applying quirks custom device configuration")
            await self._zigpy_device.apply_custom_configuration()

        # Try to add entities to claim the cluster handlers
        self._discover_new_entities()

        await asyncio.gather(
            *(endpoint.async_configure() for endpoint in self._endpoints.values())
        )

        self.emit(
            ZHA_CLUSTER_HANDLER_CFG_DONE,
            ClusterHandlerConfigurationComplete(
                device_ieee=self.ieee,
                unique_id=self.ieee,
            ),
        )

        self.debug("completed configuration")

        if (
            self.gateway.config.config.device_options.enable_identify_on_join
            and self.identify_ch is not None
            and not self.skip_configuration
        ):
            await self.identify_ch.trigger_effect(
                effect_id=Identify.EffectIdentifier.Okay,
                effect_variant=Identify.EffectVariant.Default,
            )

    def _is_entity_removed_by_quirk(self, entity: PlatformEntity) -> bool:
        if self.quirk_metadata is None:
            return False

        for meta in self.quirk_metadata.disabled_default_entities:
            _LOGGER.debug("Checking if entity %s is removed by %s", entity, meta)

            if meta.unique_id_suffix is not None and not entity.unique_id.endswith(
                meta.unique_id_suffix
            ):
                continue

            if meta.endpoint_id is not None and entity.endpoint.id != meta.endpoint_id:
                continue

            if meta.cluster_id is not None and not any(
                cluster_handler.cluster.cluster_id == meta.cluster_id
                for cluster_handler in entity.cluster_handlers.values()
            ):
                continue

            if meta.function is not None and not meta.function(entity):
                continue

            return True

        return False

    def _apply_entity_metadata_changes(self, entity: PlatformEntity) -> None:
        """Apply entity metadata changes from quirks v2."""
        if self.quirk_metadata is None:
            return

        for meta in self.quirk_metadata.changed_entity_metadata:
            if meta.unique_id_suffix is not None and not entity.unique_id.endswith(
                meta.unique_id_suffix
            ):
                continue

            if meta.endpoint_id is not None and entity.endpoint.id != meta.endpoint_id:
                continue

            if meta.cluster_id is not None and not any(
                cluster_handler.cluster.cluster_id == meta.cluster_id
                and cluster_handler.cluster.cluster_type == meta.cluster_type
                for cluster_handler in entity.cluster_handlers.values()
            ):
                continue

            if meta.function is not None and not meta.function(entity):
                continue

            # Apply metadata changes
            _LOGGER.debug(
                "Applying metadata changes from %s to entity %s", meta, entity
            )

            if meta.new_primary is not None:
                entity._attr_primary = meta.new_primary

            if meta.new_unique_id is not None:
                entity._unique_id = meta.new_unique_id

            if meta.new_translation_key is not None:
                entity._attr_translation_key = meta.new_translation_key

            if meta.new_device_class is not None:
                entity._attr_device_class = meta.new_device_class

            if meta.new_state_class is not None:
                entity._attr_state_class = meta.new_state_class

            if meta.new_entity_category is not None:
                entity._attr_entity_category = meta.new_entity_category

            if meta.new_entity_registry_enabled_default is not None:
                entity._attr_entity_registry_enabled_default = (
                    meta.new_entity_registry_enabled_default
                )

            if meta.new_fallback_name is not None:
                entity._attr_fallback_name = meta.new_fallback_name

    def _discover_new_entities(self) -> None:
        new_entities: Iterator[BaseEntity]

        if self.is_active_coordinator:
            new_entities = discovery.DEVICE_PROBE.discover_coordinator_device_entities(
                self
            )
        else:
            new_entities = discovery.DEVICE_PROBE.discover_device_entities(self)

        # Discover all applicable entities
        for entity in new_entities:
            if self._is_entity_removed_by_quirk(entity):
                continue

            # Apply any metadata changes from quirks v2
            self._apply_entity_metadata_changes(entity)

            entity.on_add()
            self._pending_entities.append(entity)

    async def async_initialize(self, from_cache: bool = False) -> None:
        """Initialize cluster handlers."""
        self.debug("started initialization")

        self._discover_new_entities()

        await self._zdo_handler.async_initialize(from_cache)
        self._zdo_handler.debug("'async_initialize' stage succeeded")

        # We intentionally do not use `gather` here! This is so that if, for example,
        # three `device.async_initialize()`s are spawned, only three concurrent requests
        # will ever be in flight at once. Startup concurrency is managed at the device
        # level.
        for endpoint in self._endpoints.values():
            try:
                await endpoint.async_initialize(from_cache)
            except Exception:  # pylint: disable=broad-exception-caught
                self.debug("Failed to initialize endpoint", exc_info=True)

        # Compute the final entities
        new_entities: dict[tuple[Platform, str], PlatformEntity] = {}

        for entity in self._pending_entities:
            entity.recompute_capabilities()

            # Ignore unsupported entities
            if not entity.is_supported() or not entity.is_supported_in_list(
                new_entities.values()
            ):
                await entity.on_remove()
                continue

            key = (entity.PLATFORM, entity.unique_id)

            # Ignore entities that already exist
            if key in new_entities:
                await entity.on_remove()
                continue

            new_entities[key] = entity

        if new_entities:
            _LOGGER.debug("Discovered new entities %r", new_entities)
            self._platform_entities.update(new_entities)

        # At this point we can compute a primary entity
        self._compute_primary_entity()

        # Sync the device's firmware version with the first platform entity
        for (platform, _unique_id), entity in self.platform_entities.items():
            if platform != Platform.UPDATE:
                continue

            assert isinstance(entity, BaseFirmwareUpdateEntity)
            self._firmware_version = entity.installed_version

            def entity_update_listener(event: EntityStateChangedEvent) -> None:
                """Listen to firmware update entity changes."""
                entity = self.get_platform_entity(event.platform, event.unique_id)
                assert isinstance(entity, BaseFirmwareUpdateEntity)
                self.async_update_firmware_version(entity.installed_version)

            self._on_remove_callbacks.append(
                entity.on_event(STATE_CHANGED, entity_update_listener)
            )

            break

        self.debug("power source: %s", self.power_source)
        self.status = DeviceStatus.INITIALIZED
        self.debug("completed initialization")

    async def on_remove(self) -> None:
        """Cancel tasks this device owns."""
        for callback in self._on_remove_callbacks:
            callback()

        for platform_entity in self._platform_entities.values():
            await platform_entity.on_remove()

        for entity in self._pending_entities:
            await entity.on_remove()

    def async_get_clusters(self) -> dict[int, dict[str, dict[int, Cluster]]]:
        """Get all clusters for this device."""
        return {
            ep_id: {
                CLUSTER_TYPE_IN: endpoint.in_clusters,
                CLUSTER_TYPE_OUT: endpoint.out_clusters,
            }
            for (ep_id, endpoint) in self._zigpy_device.endpoints.items()
            if ep_id != 0
        }

    def async_get_groupable_endpoints(self):
        """Get device endpoints that have a group 'in' cluster."""
        return [
            ep_id
            for (ep_id, clusters) in self.async_get_clusters().items()
            if Groups.cluster_id in clusters[CLUSTER_TYPE_IN]
        ]

    def async_get_std_clusters(self):
        """Get ZHA and ZLL clusters for this device."""

        return {
            ep_id: {
                CLUSTER_TYPE_IN: endpoint.in_clusters,
                CLUSTER_TYPE_OUT: endpoint.out_clusters,
            }
            for (ep_id, endpoint) in self._zigpy_device.endpoints.items()
            if ep_id != 0 and endpoint.profile_id in PROFILES
        }

    def async_get_cluster(
        self, endpoint_id: int, cluster_id: int, cluster_type: str = CLUSTER_TYPE_IN
    ) -> Cluster:
        """Get zigbee cluster from this entity."""
        clusters: dict[int, dict[str, dict[int, Cluster]]] = self.async_get_clusters()
        return clusters[endpoint_id][cluster_type][cluster_id]

    def async_get_cluster_attributes(
        self, endpoint_id, cluster_id, cluster_type=CLUSTER_TYPE_IN
    ):
        """Get zigbee attributes for specified cluster."""
        return self.async_get_cluster(endpoint_id, cluster_id, cluster_type).attributes

    def async_get_cluster_commands(
        self, endpoint_id, cluster_id, cluster_type=CLUSTER_TYPE_IN
    ):
        """Get zigbee commands for specified cluster."""
        cluster = self.async_get_cluster(endpoint_id, cluster_id, cluster_type)
        return {
            CLUSTER_COMMANDS_CLIENT: cluster.client_commands,
            CLUSTER_COMMANDS_SERVER: cluster.server_commands,
        }

    async def write_zigbee_attribute(
        self,
        endpoint_id: int,
        cluster_id: int,
        attribute: int | str,
        value: Any,
        cluster_type: str = CLUSTER_TYPE_IN,
        manufacturer: int | None = None,
    ) -> WriteAttributesResponse | None:
        """Write a value to a zigbee attribute for a cluster in this entity."""
        try:
            cluster: Cluster = self.async_get_cluster(
                endpoint_id, cluster_id, cluster_type
            )
        except KeyError as exc:
            raise ValueError(
                f"Cluster {cluster_id} not found on endpoint {endpoint_id} while"
                f" writing attribute {attribute} with value {value}"
            ) from exc

        attr_def = cluster.find_attribute(attribute)
        value = convert_zcl_value(value, attr_def.type)

        try:
            response = await cluster.write_attributes(
                {attribute: value}, manufacturer=manufacturer
            )
            self.debug(
                "set: %s for attr: %s to cluster: %s for ept: %s - res: %s",
                value,
                attribute,
                cluster_id,
                endpoint_id,
                response,
            )
            return response
        except zigpy.exceptions.ZigbeeException as exc:
            raise ZHAException(
                f"Failed to set attribute: "
                f"{ATTR_VALUE}: {value} "
                f"{ATTR_ATTRIBUTE}: {attribute} "
                f"{ATTR_CLUSTER_ID}: {cluster_id} "
                f"{ATTR_ENDPOINT_ID}: {endpoint_id}"
            ) from exc

    async def issue_cluster_command(
        self,
        endpoint_id: int,
        cluster_id: int,
        command: int,
        command_type: str,
        args: list | None,
        params: dict[str, Any] | None,
        cluster_type: str = CLUSTER_TYPE_IN,
        manufacturer: int | None = None,
    ) -> None:
        """Issue a command against specified zigbee cluster on this device."""
        try:
            cluster: Cluster = self.async_get_cluster(
                endpoint_id, cluster_id, cluster_type
            )
        except KeyError as exc:
            raise ValueError(
                f"Cluster {cluster_id} not found on endpoint {endpoint_id} while"
                f" issuing command {command} with args {args}"
            ) from exc
        commands: dict[int, ZCLCommandDef] = (
            cluster.server_commands
            if command_type == CLUSTER_COMMAND_SERVER
            else cluster.client_commands
        )
        if args is not None:
            self.warning(
                (
                    "args [%s] are deprecated and should be passed with the params key."
                    " The parameter names are: %s"
                ),
                args,
                [field.name for field in commands[command].schema.fields],
            )
            response = await getattr(cluster, commands[command].name)(*args)
        else:
            assert params is not None
            response = await getattr(cluster, commands[command].name)(
                **convert_to_zcl_values(params, commands[command].schema)
            )
        self.debug(
            "Issued cluster command: %s %s %s %s %s %s %s %s",
            f"{ATTR_CLUSTER_ID}: [{cluster_id}]",
            f"{ATTR_CLUSTER_TYPE}: [{cluster_type}]",
            f"{ATTR_ENDPOINT_ID}: [{endpoint_id}]",
            f"{ATTR_COMMAND}: [{command}]",
            f"{ATTR_COMMAND_TYPE}: [{command_type}]",
            f"{ATTR_ARGS}: [{args}]",
            f"{ATTR_PARAMS}: [{params}]",
            f"{ATTR_MANUFACTURER}: [{manufacturer}]",
        )
        if response is None:
            return  # client commands don't return a response
        if isinstance(response, Exception):
            raise ZHAException("Failed to issue cluster command") from response
        if response[1] is not ZclStatus.SUCCESS:
            raise ZHAException(
                f"Failed to issue cluster command with status: {response[1]}"
            )

    async def async_add_to_group(self, group_id: int) -> None:
        """Add this device to the provided zigbee group."""
        try:
            # A group name is required. However, the spec also explicitly states that
            # the group name can be ignored by the receiving device if a device cannot
            # store it, so we cannot rely on it existing after being written. This is
            # only done to make the ZCL command valid.
            await self._zigpy_device.add_to_group(group_id, name=f"0x{group_id:04X}")
        except (zigpy.exceptions.ZigbeeException, TimeoutError) as ex:
            self.debug(
                "Failed to add device '%s' to group: 0x%04x ex: %s",
                self._zigpy_device.ieee,
                group_id,
                str(ex),
            )

    async def async_remove_from_group(self, group_id: int) -> None:
        """Remove this device from the provided zigbee group."""
        try:
            await self._zigpy_device.remove_from_group(group_id)
        except (zigpy.exceptions.ZigbeeException, TimeoutError) as ex:
            self.debug(
                "Failed to remove device '%s' from group: 0x%04x ex: %s",
                self._zigpy_device.ieee,
                group_id,
                str(ex),
            )

    async def async_add_endpoint_to_group(
        self, endpoint_id: int, group_id: int
    ) -> None:
        """Add the device endpoint to the provided zigbee group."""
        try:
            await self._zigpy_device.endpoints[endpoint_id].add_to_group(
                group_id, name=f"0x{group_id:04X}"
            )
        except (zigpy.exceptions.ZigbeeException, TimeoutError) as ex:
            self.debug(
                "Failed to add endpoint: %s for device: '%s' to group: 0x%04x ex: %s",
                endpoint_id,
                self._zigpy_device.ieee,
                group_id,
                str(ex),
            )

    async def async_remove_endpoint_from_group(
        self, endpoint_id: int, group_id: int
    ) -> None:
        """Remove the device endpoint from the provided zigbee group."""
        try:
            await self._zigpy_device.endpoints[endpoint_id].remove_from_group(group_id)
        except (zigpy.exceptions.ZigbeeException, TimeoutError) as ex:
            self.debug(
                (
                    "Failed to remove endpoint: %s for device '%s' from group: 0x%04x"
                    " ex: %s"
                ),
                endpoint_id,
                self._zigpy_device.ieee,
                group_id,
                str(ex),
            )

    async def async_bind_to_group(
        self, group_id: int, cluster_bindings: list[ClusterBinding]
    ) -> None:
        """Directly bind this device to a group for the given clusters."""
        await self._async_group_binding_operation(
            group_id, zdo_types.ZDOCmd.Bind_req, cluster_bindings
        )

    async def async_unbind_from_group(
        self, group_id: int, cluster_bindings: list[ClusterBinding]
    ) -> None:
        """Unbind this device from a group for the given clusters."""
        await self._async_group_binding_operation(
            group_id, zdo_types.ZDOCmd.Unbind_req, cluster_bindings
        )

    async def _async_group_binding_operation(
        self,
        group_id: int,
        operation: zdo_types.ZDOCmd,
        cluster_bindings: list[ClusterBinding],
    ) -> None:
        """Create or remove a direct zigbee binding between a device and a group."""

        zdo = self._zigpy_device.zdo
        op_msg = "0x%04x: %s %s, ep: %s, cluster: %s to group: 0x%04x"
        destination_address = zdo_types.MultiAddress()
        destination_address.addrmode = uint8_t(1)
        destination_address.nwk = uint16_t(group_id)

        tasks = []

        for cluster_binding in cluster_bindings:
            if cluster_binding.endpoint_id == 0:
                continue
            if (
                cluster_binding.id
                in self._zigpy_device.endpoints[
                    cluster_binding.endpoint_id
                ].out_clusters
            ):
                op_params = (
                    self.nwk,
                    operation.name,
                    str(self.ieee),
                    cluster_binding.endpoint_id,
                    cluster_binding.id,
                    group_id,
                )
                zdo.debug(f"processing {op_msg}", *op_params)
                tasks.append(
                    (
                        zdo.request(
                            operation,
                            self.ieee,
                            cluster_binding.endpoint_id,
                            cluster_binding.id,
                            destination_address,
                        ),
                        op_msg,
                        op_params,
                    )
                )
        res = await asyncio.gather(*(t[0] for t in tasks), return_exceptions=True)
        for outcome, log_msg in zip(res, tasks):
            if isinstance(outcome, Exception):
                fmt = f"{log_msg[1]} failed: %s"
            else:
                fmt = f"{log_msg[1]} completed: %s"
            zdo.debug(fmt, *(log_msg[2] + (outcome,)))

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        msg = f"[%s](%s): {msg}"
        args = (self.nwk, self.model) + args
        _LOGGER.log(level, msg, *args, **kwargs)

    def _compute_primary_entity(self) -> None:
        """Compute the primary entity for this device."""

        # First, check if any entity is explicitly primary
        explicitly_primary = [
            entity for entity in self._platform_entities.values() if entity.primary
        ]

        if len(explicitly_primary) == 1:
            self.debug(
                "Device has a single explicitly primary entity,"
                " not performing weight matching"
            )
            return

        # It should not be possible for there to be more than one
        assert not explicitly_primary

        # For weight matching, only consider non-counter entities and entities which are
        # not explicitly marked as not primary
        candidates = [
            e
            for e in self._platform_entities.values()
            if e.enabled and hasattr(e, "info_object") and e._attr_primary is not False
        ]
        candidates.sort(reverse=True, key=lambda e: e.primary_weight)

        if not candidates:
            return

        winner = candidates[0]
        others = candidates[1:]

        # We have a clear winner
        if not others or winner.primary_weight > others[0].primary_weight:
            winner.primary = True
            del winner.info_object

            for entity in others:
                entity.primary = False
                del entity.info_object

            return

        self.debug(
            "Primary entity tie between %s and %s, no primary entity", winner, others[0]
        )

        for entity in candidates:
            entity.primary = False
            del entity.info_object

    def get_diagnostics_json(self):
        """Get ZHA device information."""

        info: dict[str, Any] = {}
        info["version"] = DIAGNOSTICS_JSON_VERSION
        info["ieee"] = str(self.ieee)
        info["nwk"] = str(self.nwk)
        info["manufacturer"] = self.device.manufacturer
        info["model"] = self.device.model
        info["friendly_manufacturer"] = self.manufacturer
        info["friendly_model"] = self.model
        info["name"] = self.name
        info["quirk_applied"] = self.quirk_applied
        info["quirk_class"] = self.quirk_class
        info["quirk_id"] = self.quirk_id
        info["manufacturer_code"] = self.manufacturer_code
        info["power_source"] = self.power_source
        info["lqi"] = self.lqi
        info["rssi"] = self.rssi
        info["last_seen"] = self.device._last_seen.isoformat()
        info["available"] = self.available
        info["device_type"] = self.device_type
        info["active_coordinator"] = self.is_active_coordinator

        node_desc = self.device.node_desc
        info["node_descriptor"] = {
            "logical_type": node_desc.logical_type.name,
            "complex_descriptor_available": bool(
                node_desc.complex_descriptor_available
            ),
            "user_descriptor_available": bool(node_desc.user_descriptor_available),
            "reserved": node_desc.reserved,
            "aps_flags": node_desc.aps_flags,
            "frequency_band": node_desc.frequency_band,
            "mac_capability_flags": node_desc.mac_capability_flags,
            "manufacturer_code": node_desc.manufacturer_code,
            "maximum_buffer_size": node_desc.maximum_buffer_size,
            "maximum_incoming_transfer_size": node_desc.maximum_incoming_transfer_size,
            "server_mask": node_desc.server_mask,
            "maximum_outgoing_transfer_size": node_desc.maximum_outgoing_transfer_size,
            "descriptor_capability_field": node_desc.descriptor_capability_field,
        }

        info["endpoints"] = {}

        for endpoint in sorted(
            self.device.non_zdo_endpoints, key=lambda ep: ep.endpoint_id
        ):
            info["endpoints"][endpoint.endpoint_id] = {
                "profile_id": endpoint.profile_id,
                "device_type": {
                    "name": (
                        (
                            PROFILES[endpoint.profile_id]
                            .DeviceType(endpoint.device_type)
                            .name
                        )
                        if endpoint.profile_id in PROFILES
                        and endpoint.device_type is not None
                        else UNKNOWN
                    ),
                    "id": endpoint.device_type,
                },
                "in_clusters": [
                    {
                        "cluster_id": f"0x{cluster_id:04x}",
                        "endpoint_attribute": cluster.ep_attribute,
                        "attributes": get_cluster_attr_data(cluster),
                    }
                    for cluster_id, cluster in sorted(endpoint.in_clusters.items())
                ],
                "out_clusters": [
                    {
                        "cluster_id": f"0x{cluster_id:04x}",
                        "endpoint_attribute": cluster.ep_attribute,
                        "attributes": get_cluster_attr_data(cluster),
                    }
                    for cluster_id, cluster in sorted(endpoint.out_clusters.items())
                ],
            }

        if isinstance(self.device, CustomDeviceV2):
            original_signature = copy.deepcopy(self.device.replacement)
        elif isinstance(self.device, zigpy.quirks.CustomDevice):
            original_signature = copy.deepcopy(self.device.signature)
        else:
            original_signature = None

        # if we have a quirked device we add the original signature to the output and
        # convert the profile_id, device_type, input_clusters and output_clusters to hex
        # representation to make it consistent with the rest of the data
        if original_signature is not None:
            if "endpoints" in original_signature:
                for ep in original_signature["endpoints"].values():
                    if "profile_id" in ep:
                        ep["profile_id"] = f"0x{ep['profile_id']:04x}"

                    if "device_type" in ep:
                        ep["device_type"] = f"0x{ep['device_type']:04x}"

                    if "input_clusters" in ep:
                        ep["input_clusters"] = [
                            f"0x{c:04x}" for c in ep["input_clusters"]
                        ]

                    if "output_clusters" in ep:
                        ep["output_clusters"] = [
                            f"0x{c:04x}" for c in ep["output_clusters"]
                        ]

            if "node_desc" in original_signature:
                original_signature["node_desc"] = original_signature[
                    "node_desc"
                ].as_dict()

            info["original_signature"] = original_signature

        info["zha_lib_entities"] = defaultdict(list)

        for (platform, _unique_id), platform_entity in sorted(
            self.platform_entities.items()
        ):
            info_object = dataclasses.asdict(platform_entity.info_object)
            info_object["cluster_handlers"].sort(key=lambda i: i["unique_id"])
            info_object["migrate_unique_ids"] = list(info_object["migrate_unique_ids"])
            info_object["device_ieee"] = str(info_object["device_ieee"])

            for cluster_handler_info in info_object["cluster_handlers"]:
                cluster_info = cluster_handler_info["cluster"]

                if cluster_info is not None:
                    cluster_info.pop("commands", None)

            obj: dict[str, Any] = {
                "info_object": info_object,
                "state": platform_entity.state,
            }

            if platform_entity.extra_state_attribute_names is not None:
                obj["extra_state_attributes"] = sorted(
                    platform_entity.extra_state_attribute_names
                )

            info["zha_lib_entities"][platform].append(obj)

        topology = self.gateway.application_controller.topology
        info["neighbors"] = [
            {
                "device_type": neighbor.device_type.name,
                "rx_on_when_idle": neighbor.rx_on_when_idle.name,
                "relationship": neighbor.relationship.name,
                "extended_pan_id": str(neighbor.extended_pan_id),
                "ieee": str(neighbor.ieee),
                "nwk": str(neighbor.nwk),
                "permit_joining": neighbor.permit_joining.name,
                "depth": neighbor.depth,
                "lqi": neighbor.lqi,
            }
            for neighbor in topology.neighbors[self.device.ieee]
        ]

        info["routes"] = [
            {
                "dest_nwk": str(route.DstNWK),
                "route_status": str(route.RouteStatus.name),
                "memory_constrained": bool(route.MemoryConstrained),
                "many_to_one": bool(route.ManyToOne),
                "route_record_required": bool(route.RouteRecordRequired),
                "next_hop": str(route.NextHop),
            }
            for route in topology.routes[self.device.ieee]
        ]

        return info
